#! python -v
# -*- coding: utf-8 -*-

from refal.lexer_refal import *
from refal.parser_refal import *
from refal.algorithm import *
from refal.constants import *

import sys
import argparse
import os


def get_file_name(file_path):
    head, tail = split(file_path)
    return tail


def new_parser():
    parser = argparse.ArgumentParser(description="Arguments for program refalcheck")
    parser.add_argument("file", help="File for which format functions is calculated")
    parser.add_argument("file_type", nargs="*", help="Files, contained required format functions")
    return parser


def calculate_ast(file_path, is_file_type=False):
    try:
        file_refal = open(file_path, "r+", encoding="utf-8")
    except (FileNotFoundError, IOError) as e:
        sys.stderr.write(str(e))
        sys.exit(-1)

    if DEBUG_MODE:
        print("Lexer %s" % get_file_name(file_path))

    program = file_refal.read()
    lexer = Lexer(program)

    cur_token = lexer.next_token()
    list_token = [cur_token]

    if DEBUG_MODE:
        print(cur_token)
    while cur_token.tag != DomainTag.Eop:
        cur_token = lexer.next_token()
        if DEBUG_MODE:
            print(cur_token)
        list_token.append(cur_token)

    list_token = refactor_char_token(list_token)

    file_refal.close()

    if DEBUG_MODE:
        print(LINE_DELIMITER)

    if DEBUG_MODE:
        print("Parser %s" % get_file_name(file_path))

    if not is_file_type:
        parser = ParserRefal(list_token)
    else:
        parser = ParserRefalType(list_token)
    parser.parse()

    parser.semantics()

    if DEBUG_MODE:
        print(LINE_DELIMITER)

    return parser


def append_ast(parser_left, parser_right):
    if not parser_left.isError and not parser_right.isError:
        parser_left.ast.functions.extend(parser_right.ast.functions)
        return parser_left
    else:
        sys.stderr.write("Can't append AST, because it was error")


def main():
    arg_parse = new_parser()
    args = arg_parse.parse_args()

    file_refal_path = None
    if DEBUG_MODE:
        file_refal_path = join(TEST_DIRECTORY, args.file).replace("\\", "/")
    else:
        file_refal_path = args.file

    parser_refal = calculate_ast(file_refal_path)

    file_type_path = os.path.join(os.path.dirname(__file__), "built-in.type")
    parser_refal_type = calculate_ast(file_type_path, True)

    for file in args.file_type:
        if DEBUG_MODE:
            file_refal_path = join(TEST_DIRECTORY, file).replace("\\", "/")
        else:
            file_refal_path = file

        append_ast(parser_refal_type, calculate_ast(file_refal_path, True))

    parser_refal.semantics_call(parser_refal_type.ast)

    if not parser_refal.isError and not parser_refal_type.isError:
        calculation = Calculation(parser_refal.ast, parser_refal_type.ast)
        calculation.calculate_system()
        print(calculation.print_format_function())


if __name__ == "__main__":
    main()



#! python -v
# -*- coding: utf-8 -*-


def generate_index():
    generate_index.counter += 1
    return generate_index.counter


generate_index.counter = -1


def generate_built_in_index():
    generate_built_in_index.counter -= 1
    return generate_built_in_index.counter


generate_built_in_index.counter = 0



#! python -v
# -*- coding: utf-8 -*-

from refal.ast import *
from refal.constants import *
from refal.utils import *

from copy import deepcopy

import math
import sys


def change_variable_index(variable, index):
    if isinstance(variable, Variable):
        variable.index = index
        return variable
    else:
        raise ValueError("variable must be instance of class Variable (s.Index, t.Index, e.Index)")


def change_variable_index_and_index_sentence(variable, index, sentence_index):
    if isinstance(variable, Variable):
        variable.index = index
        variable.sentence_index = sentence_index
        return variable
    else:
        raise ValueError("variable must be instance of class Variable (s.Index, t.Index, e.Index)")


def is_hard_term(term):
    if isinstance(term, Variable):
        if term.type_variable == Type.e:
            return False
        else:
            return True
    elif isinstance(term, CallBrackets):
        return True
    else:
        return True


def is_symbol(term):
    if isinstance(term, (CompoundSymbol, Macrodigit, Char)):
        return True
    else:
        return False


def calc_complexity(terms):
    complexity = 0
    for term in terms:
        if isinstance(term, Variable) and term.type_variable == Type.s:
            complexity += 2
        elif isinstance(term, Variable) and term.type_variable == Type.t:
            complexity += 1
        elif isinstance(term, Variable) and term.type_variable == Type.e:
            complexity -= 1
        elif is_symbol(term):
            complexity += 3
        elif isinstance(term, StructuralBrackets):
            complexity += 2
        elif isinstance(term, SpecialVariable) and term.type_variable == SpecialType.none:
            complexity += math.inf
    return complexity + 1


def append_char_term(terms):
    i = 0
    append_terms = []
    while i < len(terms):
        if isinstance(terms[i], Char):
            j = i + 1
            value = terms[i].value
            while j < len(terms):
                if not isinstance(terms[j], Char):
                    append_terms.append(Char("'" + value.replace("'", "") + "'"))
                    value = ""
                    i = j - 1
                    break
                else:
                    value += terms[j].value
                    j += 1
            if value != "" and j == len(terms):
                append_terms.append(Char("'" + value.replace("'", "") + "'"))
                i = j
        elif isinstance(terms[i], CallBrackets):
            append_terms.append(CallBrackets(terms[i].value, terms[i].pos, append_char_term(terms[i].content)))
        elif isinstance(terms[i], StructuralBrackets):
            append_terms.append(StructuralBrackets(append_char_term(terms[i].value)))
        else:
            append_terms.append(terms[i])
        i += 1
    return append_terms


class SpecialType(Enum):
    in_function = 1
    out_function = 2
    none = 3


class SpecialVariable(Term):

    def __init__(self, value, type_variable, sentence_index=-1):
        super(SpecialVariable, self).__init__(value)
        self.type_variable = type_variable
        self.sentence_index = sentence_index

    def __eq__(self, other):
        return isinstance(other, SpecialVariable) \
               and self.value == other.value \
               and self.type_variable == other.type_variable

    def __str__(self):
        if self.type_variable == SpecialType.in_function:
            return "in[" + str(self.value) + "]"
        elif self.type_variable == SpecialType.out_function:
            return "out[" + str(self.value) + "]"
        else:
            return super(SpecialVariable, self).__str__()


class EqType(Enum):
    Term = 1
    Expr = 2


class Equation(object):

    def __init__(self, left_part, right_part, type_equation, func_name, index_sentence, sentence):
        self.left_part = left_part
        self.right_part = right_part
        self.type_equation = type_equation

        self.func_name = func_name
        self.index_sentence = index_sentence
        self.sentence = sentence

    def __eq__(self, other):
        return isinstance(other, Equation) \
               and self.left_part == other.left_part \
               and self.right_part == other.right_part \
               and self.type_equation == other.type_equation

    def __str__(self):
        return str(self.left_part) + " : " + str(self.right_part)

    def __hash__(self):
        return hash(self.__str__())


class Substitution(object):

    def __init__(self, left_part, right_part, alternative_right_part=None):
        self.left_part = left_part
        self.right_part = right_part

        self.alternative_right_part = alternative_right_part

    def __hash__(self):
        return hash(self.__str__())

    def __eq__(self, other):
        return isinstance(other, Substitution) \
               and self.left_part == other.left_part \
               and self.right_part == other.right_part

    def __str__(self):
        output_str = str(self.left_part) + " -> " + str(self.right_part)
        if self.alternative_right_part is not None:
            output_str += " | " + str(self.alternative_right_part)
        return output_str


class Assignment(object):

    def __init__(self, left_part, right_part):
        self.left_part = left_part
        self.right_part = right_part

    def __hash__(self):
        return hash(self.__str__())

    def __eq__(self, other):
        return isinstance(other, Assignment) \
               and self.left_part == other.left_part \
               and self.right_part == other.right_part

    def __str__(self):
        return str(self.left_part) + " <- " + str(self.right_part)


def make_alternative(substitutions):
    i = 0
    substitutions_result = []
    substitutions_variant = []
    while i < len(substitutions):
        j = 0
        while j < len(substitutions_variant):
            if substitutions_variant[j].left_part == substitutions[i].left_part:
                break
            j += 1
        if j == len(substitutions_variant):
            substitutions_variant.append(substitutions[i])
        else:
            substitutions_result.append(deepcopy(substitutions_variant))
            substitutions_variant = substitutions_variant[:j]
            i -= 1
        i += 1
    substitutions_result.append(substitutions_variant)
    return substitutions_result


def has_alternative(substitutions):
    for subst in substitutions:
        if subst.alternative_right_part is not None:
            return True
    return False


def get_alternative_substitution(substitutions):
    for subst in substitutions:
        if subst.alternative_right_part is not None:
            return subst
    return None


def get_substitution_function_format(ast):
    format_substitution_function = []

    for function in ast.functions:
        if isinstance(function, Definition):
            for sentence in function.sentences:
                if not sentence.no_substitution:
                    format_substitution_function.append(Substitution(
                        Expression([SpecialVariable(function.name, SpecialType.in_function)]),
                        sentence.pattern))

                    if not any(isinstance(term, SpecialVariable) and term.type_variable == SpecialType.out_function
                               for term in sentence.result.terms):
                        format_substitution_function.append(Substitution(
                            Expression([SpecialVariable(function.name, SpecialType.out_function)]),
                            sentence.result))

    return format_substitution_function


def has_repeated_variable(exprs):
    i, j = 0, 0
    while i < len(exprs.terms):
        while j < len(exprs.terms):
            if i == j:
                j += 1
                continue
            else:
                if isinstance(exprs.terms[i], Variable) and isinstance(exprs.terms[j], Variable) \
                        and exprs.terms[i] == exprs.terms[j]:
                    return True
                else:
                    j += 1
        i += 1
    return False


def delete_same_assignments(assignments, substitution):
    i, j = 0, 0
    while i < len(assignments):
        while j < len(substitution):

            if assignments[i].right_part == substitution[j].left_part:
                del assignments[i]
                i -= 1
                break

            j += 1
        i += 1

    return assignments, substitution


class Calculation(object):

    def __init__(self, ast, ast_type):
        self.ast = deepcopy(ast)
        self.ast_type = deepcopy(ast_type)

        self.block_to_condition()
        self.numerate_variable()

        if DEBUG_MODE:
            print(self.ast)
            print(LINE_DELIMITER)
            print(self.ast_type)

        self.default_format_function = self.prepare_default_format_function()
        self.format_function = []

        self.system_result = list(dict.fromkeys(self.create_equation_result()))
        self.system_condition_result, self.system_condition = self.create_equation_condition()

        if DEBUG_MODE:
            print(LINE_DELIMITER)
            for equation in self.system_result:
                print(equation)
            for equation in self.system_condition_result:
                print(equation)
            for equation in self.system_condition:
                print(equation)
            print(LINE_DELIMITER)

        self.error_messages = []

    def get_call(self, term):
        if isinstance(term, StructuralBrackets):
            term_call = []
            for terms in term.value:
                term_call.extend(self.get_call(terms))
            return term_call
        elif isinstance(term, CallBrackets):
            term_call = [term.value]
            for terms in term.content:
                term_call.extend(self.get_call(terms))
            return term_call
        elif isinstance(term, Variable):
            return []
        else:
            return []

    def contains_call(self, term):
        if isinstance(term, StructuralBrackets):
            return any(self.contains_call(term) for term in term.value)
        elif isinstance(term, CallBrackets):
            return True
        elif isinstance(term, Variable):
            return False
        else:
            return False

    def get_variables(self, term):
        if isinstance(term, StructuralBrackets):
            variable = []
            for term in term.value:
                variable.extend(self.get_variables(term))
            return variable
        elif isinstance(term, CallBrackets):
            variable = []
            for term in term.content:
                variable.extend(self.get_variables(term))
            return variable
        elif isinstance(term, Variable):
            return [term]
        else:
            return []

    def get_none(self, term):
        if isinstance(term, StructuralBrackets):
            for term in term.value:
                none_term = self.get_none(term)
                if none_term is not None:
                    return none_term
        elif isinstance(term, SpecialVariable) and term.type_variable == SpecialType.none:
            return term
        elif isinstance(term, Expression):
            for term in term.terms:
                none_term = self.get_none(term)
                if none_term is not None:
                    return none_term
        else:
            return None

    def numerate_variable_function(self, exprs):
        variable_dict = dict()
        for term in exprs:
            variables = self.get_variables(term)
            for i in range(len(variables)):
                if variables[i].value not in variable_dict:
                    if variables[i].index >= 0:
                        variable_dict[variables[i].value] = generate_index()
                    else:
                        variable_dict[variables[i].value] = generate_built_in_index()
                variables[i] = change_variable_index(variables[i], variable_dict[variables[i].value])
        return exprs

    def numerate_variable(self):
        index = 0
        for function in self.ast.functions:
            if isinstance(function, Definition):
                self.numerate_variable_rec(function.sentences, index, dict())
        for function_type in self.ast_type.functions:
            if isinstance(function_type, DefinitionType):
                self.numerate_variable_built_in(function_type)

    def numerate_variable_rec(self, sentences, index, variable_dict):
        for sentence in sentences:
            index += 1
            for term in sentence.pattern.terms:
                variables = self.get_variables(term)
                for i in range(len(variables)):
                    if variables[i].value not in variable_dict:
                        variable_dict[variables[i].value] = generate_index()
                    variables[i] = change_variable_index_and_index_sentence(variables[i], variable_dict[variables[i].value], index)
            for condition in sentence.conditions:
                for term_condition in condition.pattern.terms:
                    variables = self.get_variables(term_condition)
                    for i in range(len(variables)):
                        if variables[i].value not in variable_dict:
                            variable_dict[variables[i].value] = generate_index()
                        variables[i] = change_variable_index_and_index_sentence(variables[i], variable_dict[variables[i].value],
                                                                                index)
                for term_result in condition.result.terms:
                    variables = self.get_variables(term_result)
                    for i in range(len(variables)):
                        if variables[i].value not in variable_dict:
                            variable_dict[variables[i].value] = generate_index()
                        variables[i] = change_variable_index_and_index_sentence(variables[i], variable_dict[variables[i].value],
                                                                                index)
            if sentence.block:
                self.numerate_variable_rec(sentence.block, 0, variable_dict)
            for term_result in sentence.result.terms:
                variables = self.get_variables(term_result)
                for i in range(len(variables)):
                    if variables[i].value not in variable_dict:
                        variable_dict[variables[i].value] = generate_index()
                    variables[i] = change_variable_index_and_index_sentence(variables[i], variable_dict[variables[i].value], index)

    def numerate_variable_built_in(self, function_type):
        for term in function_type.pattern.terms:
            variables = self.get_variables(term)
            for i in range(len(variables)):
                variables[i] = change_variable_index_and_index_sentence(variables[i], generate_built_in_index(), 0)
        for term in function_type.result.terms:
            variables = self.get_variables(term)
            for i in range(len(variables)):
                variables[i] = change_variable_index_and_index_sentence(variables[i], generate_built_in_index(), 0)

    def make_hard(self, expr):
        if len(expr.terms) == 0:
            return Expression([])
        elif is_symbol(expr.terms[0]):
            return Expression([expr.terms[0], *self.make_hard(Expression(expr.terms[1:])).terms])
        elif is_symbol(expr.terms[-1]):
            return Expression([*self.make_hard(Expression(expr.terms[:-1])).terms, expr.terms[-1]])
        elif isinstance(expr.terms[0], Variable) and (
                expr.terms[0].type_variable == Type.s or expr.terms[0].type_variable == Type.t):
            return Expression([expr.terms[0], *self.make_hard(Expression(expr.terms[1:])).terms])
        elif isinstance(expr.terms[-1], Variable) and \
                (expr.terms[-1].type_variable == Type.s or expr.terms[-1].type_variable == Type.t):
            return Expression([*self.make_hard(Expression(expr.terms[:-1])).terms, expr.terms[-1]])
        elif isinstance(expr.terms[0], StructuralBrackets):
            return Expression([StructuralBrackets(*self.make_hard(Expression(expr.terms[0].value)).terms),
                               *self.make_hard(Expression(expr.terms[1:])).terms])
        elif isinstance(expr.terms[-1], StructuralBrackets):
            return Expression([*self.make_hard(Expression(expr.terms[:-1])).terms,
                               StructuralBrackets(*self.make_hard(Expression(expr.terms[-1].value)).terms)])
        elif len(expr.terms) == 1 and isinstance(expr.terms[0], Variable) and expr.terms[0].type_variable == Type.e:
            return Expression([expr.terms[0]])
        elif len(expr.terms) == 1 and isinstance(expr.terms[0], SpecialVariable) \
                and expr.terms[0].type_variable == SpecialType.none:
            return Expression([expr.terms[0]])
        else:
            index = generate_index()
            e_variable = Variable("%d".format(index), Type.e, None, index)
            return Expression([e_variable])

    def prepare_default_format_function(self):

        format_function = []

        call_functions = []

        for function in self.ast.functions:
            if isinstance(function, Definition):
                for sentence in function.sentences:
                    for condition in sentence.conditions:
                        for term_result in condition.result.terms:
                            call_functions.extend(self.get_call(term_result))

                    for term_result in sentence.result.terms:
                        call_functions.extend(self.get_call(term_result))

        call_functions = list(set(call_functions))

        for definition_type_function in self.ast_type.functions:
            if isinstance(definition_type_function, DefinitionType):
                for call_function in call_functions:
                    if definition_type_function.name == call_function:
                        in_format = Expression(
                            [SpecialVariable(definition_type_function.name, SpecialType.in_function)])
                        out_format = Expression(
                            [SpecialVariable(definition_type_function.name, SpecialType.out_function)])
                        format_function.append(Substitution(in_format, definition_type_function.pattern))
                        format_function.append(Substitution(out_format, definition_type_function.result))

        return format_function

    def create_equation_condition(self):
        all_equations = []
        equations_conditions = []
        for function in self.ast.functions:
            if isinstance(function, Definition):
                equations = []
                for sentence in function.sentences:
                    for condition in sentence.conditions:
                        for i in range(len(condition.result.terms)):
                            if isinstance(condition.result.terms[i], CallBrackets):
                                sentence.has_call = True
                                equations_new, _1, _2 = self.replace_call(condition.result.terms[i].content,
                                                                          condition.result.terms[i].value,
                                                                          function.sentences.index(sentence),
                                                                          sentence)
                                equations.extend(equations_new)
                                condition.result.terms[i] = SpecialVariable(condition.result.terms[i].value,
                                                                            SpecialType.out_function)
                            elif isinstance(condition.result.terms[i], StructuralBrackets):
                                equations_new, _1, _2 = self.replace_call(condition.result.terms[i].value,
                                                                          None,
                                                                          function.sentences.index(sentence),
                                                                          sentence)
                                equations.extend(equations_new)

                        equations_conditions.append(Equation(condition.result, condition.pattern, EqType.Expr,
                                                             function.name, function.sentences.index(sentence),
                                                             sentence))
                all_equations.extend(equations)
        return all_equations, equations_conditions

    def create_equation_result(self):
        all_equations = []
        for function in self.ast.functions:
            if isinstance(function, Definition):
                equations = []
                for sentence in function.sentences:
                    for i in range(len(sentence.result.terms)):
                        if isinstance(sentence.result.terms[i], CallBrackets):
                            sentence.has_call = True
                            equations_new, _1, _2 = self.replace_call(sentence.result.terms[i].content,
                                                                      sentence.result.terms[i].value,
                                                                      function.sentences.index(sentence),
                                                                      sentence)
                            equations.extend(equations_new)
                            sentence.result.terms[i] = SpecialVariable(sentence.result.terms[i].value,
                                                                       SpecialType.out_function)
                        elif isinstance(sentence.result.terms[i], StructuralBrackets):
                            equations_new, _1, _2 = self.replace_call(sentence.result.terms[i].value,
                                                                      None,
                                                                      function.sentences.index(sentence),
                                                                      sentence)
                            equations.extend(equations_new)
                in_format = Expression([SpecialVariable(function.name, SpecialType.in_function)])
                out_format = Expression([SpecialVariable(function.name, SpecialType.out_function)])

                found_default_in_format = False
                found_default_out_format = False
                for definition_type_function in self.default_format_function:
                    if definition_type_function.left_part == in_format:
                        found_default_in_format = True
                    if definition_type_function.right_part == out_format:
                        found_default_out_format = True
                if not found_default_in_format:
                    self.format_function.append(
                        Substitution(Expression([SpecialVariable(function.name, SpecialType.in_function)]),
                                     Expression([SpecialVariable("@", SpecialType.none)])))
                if not found_default_out_format:
                    self.format_function.append(
                        Substitution(Expression([SpecialVariable(function.name, SpecialType.out_function)]),
                                     Expression([SpecialVariable("@", SpecialType.none)])))
                all_equations.extend(equations)
        return all_equations

    def block_to_condition(self):
        for function in self.ast.functions:
            if isinstance(function, Definition):
                for sentence in function.sentences:
                    if sentence.block:
                        function.sentences.remove(sentence)

                        for sentence_block in sentence.block:
                            sentence_copy = deepcopy(sentence)
                            sentence_copy.block = []

                            sentence_result = sentence_copy.result
                            sentence_copy.conditions.append(Condition(sentence_result, sentence_block.pattern))
                            sentence_copy.result = sentence_block.result

                            function.sentences.append(sentence_copy)

    def replace_call(self, terms, func_name, index_sentence, sentence):
        equations = []
        i = 0
        while i < len(terms):
            if isinstance(terms[i], CallBrackets):
                sentence.has_call = True
                equations_new, _, flag = self.replace_call(terms[i].content, terms[i].value,
                                                           index_sentence, sentence)
                terms[i] = SpecialVariable(terms[i].value, SpecialType.out_function)
                equations.extend(equations_new)
            elif isinstance(terms[i], StructuralBrackets):
                equations_new, terms_new, flag = self.replace_call(terms[i].value, None, index_sentence,
                                                                   sentence)
                terms[i] = StructuralBrackets(terms_new)
                if flag:
                    equations.extend(equations_new)
            i += 1
        if func_name is not None:
            in_variable = Expression([SpecialVariable(func_name, SpecialType.in_function)])
            return [Equation(Expression(terms), in_variable, EqType.Expr, func_name, index_sentence, sentence),
                    *equations], terms, True
        else:
            return [], terms, False

    def apply_substitution(self, expr, subst, flag=False):
        res = []
        i = 0
        exprs = deepcopy(expr)
        while i < len(exprs.terms):
            j = i + 1
            exists_substitution = False
            while j <= len(exprs.terms):
                sub_expr = Expression(exprs.terms[i:j])
                for substitution in subst:
                    if substitution.left_part == sub_expr:
                        exists_substitution = True
                        if len(substitution.left_part.terms) == 1 and isinstance(substitution.left_part.terms[0],
                                                                                 SpecialVariable) and not flag:

                            res.extend(self.numerate_variable_function(
                                self.apply_substitution(substitution.right_part, subst, flag).terms
                            ))

                        else:
                            res.extend(self.apply_substitution(substitution.right_part, subst, flag).terms)
                        del exprs.terms[i:j]
                        break
                if exists_substitution:
                    break
                j += 1
            if exists_substitution:
                continue
            if isinstance(exprs.terms[i], (Variable, SpecialVariable)):
                exists_substitution = False
                for substitution in subst:
                    if substitution.left_part == exprs or \
                            (len(substitution.left_part.terms) == 1 and substitution.left_part.terms[0] == exprs.terms[
                                i]):
                        exists_substitution = True
                        if isinstance(substitution.left_part.terms[0], SpecialVariable):
                            res.extend(self.numerate_variable_function(
                                self.apply_substitution(substitution.right_part, subst, flag).terms
                            ))
                        else:
                            res.extend(self.apply_substitution(substitution.right_part, subst, flag).terms)
                if not exists_substitution:
                    res.append(exprs.terms[i])
            elif isinstance(exprs.terms[i], StructuralBrackets):
                res.extend([StructuralBrackets(self.apply_substitution(Expression(exprs.terms[i].value), subst, flag).terms)])
            else:
                res.append(exprs.terms[i])
            i += 1
        return Expression(res)

    def apply_assignment(self, expr, assignments):
        res = []
        i = 0
        exprs = deepcopy(expr)
        while i < len(exprs.terms):
            if isinstance(exprs.terms[i], Variable):
                exist_assignment = False
                for assign in assignments:
                    if len(assign.right_part.terms) == 1 and assign.right_part.terms[0] == exprs.terms[i]:
                        exist_assignment = True
                        res.extend(assign.left_part.terms)
                if not exist_assignment:
                    res.append(exprs.terms[i])
            elif isinstance(exprs.terms[i], StructuralBrackets):
                res.extend([StructuralBrackets(self.apply_assignment(Expression(exprs.terms[i].value), assignments)
                                               .terms)])
            else:
                res.append(exprs.terms[i])
            i += 1
        return Expression(res)

    def generalization_expression(self, format_substitution_function):
        group_generalization = dict()
        for subst in [*format_substitution_function, *self.default_format_function]:
            if subst.left_part not in group_generalization:
                group_generalization[subst.left_part] = [deepcopy(subst.right_part)]
            else:
                group_generalization[subst.left_part].append(deepcopy(subst.right_part))

        format_functions = []
        for key_group in group_generalization.keys():
            format_function = self.generalization(group_generalization[key_group])
            format_functions.append(Substitution(key_group, format_function))

        format_functions_result = [format_function for format_function in format_functions
                                   if len(format_function.left_part.terms) > 0 and
                                   isinstance(format_function.left_part.terms[0], SpecialVariable) and
                                   format_function.left_part.terms[0].type_variable != SpecialType.none]
        return format_functions_result

    def calculate_equation(self, system, ast):
        substitutions = []
        assignments = []
        k = 0

        while k < len(system):
            eq = deepcopy(system[k])

            eq.left_part = self.apply_substitution(eq.left_part, [*substitutions, *self.format_function,
                                                                  *self.default_format_function])
            eq.right_part = self.apply_substitution(eq.right_part, [*substitutions, *self.format_function,
                                                                    *self.default_format_function])

            if not (len(eq.right_part.terms) > 0 and
                    any(isinstance(term, SpecialVariable) and term.type_variable == SpecialType.none
                        for term in eq.right_part.terms)) \
                    and not (len(eq.left_part.terms) > 0 and
                             any(isinstance(term, SpecialVariable) and term.type_variable == SpecialType.none
                                 for term in eq.left_part.terms)):
                status, substitution, assignment, _, _ = self.match_equation(eq, substitutions, assignments,
                                                                             system)

                assignment, substitution = delete_same_assignments(assignment, substitution)

                if has_alternative(substitution):
                    status, substitution_result, assignments_result, k = self.calculate_alternative_system(system,
                                                                                                        substitution,
                                                                                                        assignment,
                                                                                                        system[k], k,
                                                                                                        ast)

                if status != "Failure":
                    substitution = [subst for subst in substitution if not subst.left_part == subst.right_part]
                    assignment = [assign for assign in assignment if not assign.left_part == assign.right_part]

                    for subst in substitution:
                        if subst not in substitutions:
                            substitutions.append(subst)

                    for assign in assignment:
                        if assign not in assignments:
                            assignments.append(assign)

                if status == "Failure":
                    error_message = "Function %s, sentence %d, there isn't solution for equation: %s" % (
                        system[k].func_name, system[k].index_sentence, self.print_equation(system[k]))

                    if not any(msg.startswith(error_message) for msg in self.error_messages):
                        substitution = self.apply_substitution(system[k].right_part,
                                                               [*substitution,
                                                                *substitutions,
                                                                *self.format_function,
                                                                *self.default_format_function])

                        error_message = error_message + " => " + self.print_expr(substitution) + "\n"
                        self.error_messages.append(error_message)

                    for function in ast.functions:
                        if isinstance(function, Definition):
                            for sentence in function.sentences:
                                if system[k].sentence == sentence:
                                    sentence.no_substitution = True

            else:
                substitution = [Substitution(eq.left_part, eq.right_part)]
                substitution = [subst for subst in substitution if not subst.left_part == subst.right_part]

                for subst in substitution:
                    if subst not in substitutions:
                        substitutions.append(subst)
            k += 1
        return substitutions, assignments

    def calculate_alternative_system(self, system, substitutions, assignments, equation, j, ast):
        substitution_alternative = get_alternative_substitution(substitutions)

        if DEBUG_MODE:
            print(LINE_DELIMITER)
            print(substitution_alternative)
            print(LINE_DELIMITER)

        substitutions.remove(substitution_alternative)

        substitution_first_variant = Substitution(substitution_alternative.left_part,
                                                  substitution_alternative.right_part)

        status_first, substitutions_first, assignments_first, index_first = self.calculate_alternative_system_rec(system,
                                                                                                     [
                                                                                                         substitution_first_variant,
                                                                                                         *substitutions],
                                                                                                     assignments, j,
                                                                                                     ast)

        substitution_second_variant = Substitution(substitution_alternative.left_part,
                                                   substitution_alternative.alternative_right_part)

        status_second, substitutions_second, assignments_second, index_second = self.calculate_alternative_system_rec(system,
                                                                                                        [
                                                                                                            substitution_second_variant,
                                                                                                            *substitutions],
                                                                                                        assignments, j,
                                                                                                        ast)
        index_result = None
        if status_first == "Success":
            for subst in substitutions_first:
                if subst not in substitutions:
                    substitutions.append(subst)
            if substitutions_first not in substitutions:
                substitutions.append(substitution_first_variant)

            for assign in assignments_first:
                if assign not in assignments:
                    assignments.append(assign)

            index_result = index_first

        if status_second == "Success":
            for subst in substitutions_second:
                if subst not in substitutions:
                    substitutions.append(subst)
            if substitutions_second not in substitutions:
                substitutions.append(substitution_second_variant)

            for assign in assignments_second:
                if assign not in assignments:
                    assignments.append(assign)

            index_result = index_second

        if status_first == "Failure" and status_second == "Failure":

            index_result = max(index_first, index_second)

            error_message = "Function %s, sentence %d, there isn't solution for equation: %s" % (
                system[index_result].func_name, system[index_result].index_sentence, self.print_equation(system[index_result]))

            if not any(msg.startswith(error_message) for msg in self.error_messages):
                substitution = self.apply_substitution(system[index_result].right_part,
                                                       [*substitutions,
                                                        *self.format_function,
                                                        *self.default_format_function])

                error_message = error_message + " => " + self.print_expr(substitution) + "\n"
                self.error_messages.append(error_message)

                for function in ast.functions:
                    if isinstance(function, Definition):
                        for sentence in function.sentences:
                            if system[j].sentence == sentence:
                                sentence.no_substitution = True

            return "Failure", [], [], index_result

        if DEBUG_MODE:
            print(LINE_DELIMITER)
            for subst in substitutions:
                print(subst)
            for assign in assignments:
                print(assign)
            print(LINE_DELIMITER)

        return "Success", list(dict.fromkeys(substitutions)), list(dict.fromkeys(assignments)), index_result

    def calculate_alternative_system_rec(self, system, substitutions, assignments, j, ast):
        substitutions_copy = deepcopy(substitutions)
        assignments_copy = deepcopy(assignments)

        is_first = True
        while j < len(system):
            eq = deepcopy(system[j])
            eq.left_part = self.apply_substitution(eq.left_part, [*substitutions_copy,
                                                                  *self.format_function,
                                                                  *self.default_format_function
                                                                  ], is_first)
            eq.right_part = self.apply_substitution(eq.right_part,
                                                    [*substitutions_copy,
                                                     *self.format_function,
                                                     *self.default_format_function
                                                     ], is_first)

            status, substitution_result, assignments_result, _, _ = self.match_equation(eq, substitutions_copy,
                                                                                        assignments_copy,
                                                                                        self.system_result)

            assignments_result, substitution_result = delete_same_assignments(assignments_result, substitution_result)

            if status == "Failure":
                return status, [], [], j

            if has_alternative(substitution_result):
                substitution_all = []
                for subst in substitution_result:
                    if subst not in substitution_all:
                        substitution_all.append(subst)

                for subst in substitutions_copy:
                    if subst not in substitution_all:
                        substitution_all.append(subst)

                assignments_all = []
                for assign in assignments_result:
                    if assign not in assignments_all:
                        assignments_all.append(assign)

                for assign in assignments_copy:
                    if assign not in assignments_all:
                        assignments_all.append(assign)

                status, substitution_alternative, assignments_alternative, index = \
                    self.calculate_alternative_system(system, substitution_all, assignments_all, system[j],
                                                      j, ast)

                if status == "Failure":
                    return status, [], [], index

                for subst in substitution_alternative:
                    if subst not in substitutions_copy:
                        substitutions_copy.append(subst)

                for assign in assignments_alternative:
                    if assign not in assignments_copy:
                        assignments_copy.append(assign)

                j = index
            else:
                substitution_result = [subst for subst in substitution_result if
                                       not subst.left_part == subst.right_part]

                assignments_result = [assign for assign in assignments_result if
                                      not assign.left_part == assign.right_part]

                for subst in substitution_result:
                    if subst not in substitutions_copy:
                        substitutions_copy.append(subst)

                for assign in assignments_result:
                    if assign not in assignments_copy:
                        assignments_copy.append(assign)

            j += 1
            is_first = False

        return "Success", substitutions_copy, assignments_copy, j

    def calculate_system(self):
        while True:
            substitutions = []
            assignments = []

            ast = deepcopy(self.ast)

            if len(self.system_condition_result) > 0:
                substitutions_condition, assignments_condition = self.calculate_equation(self.system_condition_result,
                                                                                         ast)

                substitutions_variant = make_alternative(substitutions_condition)
                for substitution_var in substitutions_variant:
                    substitution_var.extend(substitutions)

                substitutions = substitutions_variant

                assignments.extend(assignments_condition)

                for function in ast.functions:
                    if isinstance(function, Definition):
                        sentence_result = []
                        while len(function.sentences) > 0:
                            j = 0
                            if len(substitutions) > 0:
                                new_sentence = deepcopy(function.sentences[0])
                                del function.sentences[0]
                                while j < len(substitutions):

                                    sentence_copy = deepcopy(new_sentence)

                                    for condition in sentence_copy.conditions:
                                        condition.pattern = self.apply_substitution(condition.pattern,
                                                                                    [*substitutions[j],
                                                                                     *self.format_function,
                                                                                     *self.default_format_function])
                                        condition.pattern = self.make_hard(
                                            self.apply_assignment(condition.pattern, assignments))

                                        condition.result = self.apply_substitution(condition.result,
                                                                                   [*substitutions[j],
                                                                                    *self.format_function,
                                                                                    *self.default_format_function])
                                        condition.result = self.apply_assignment(condition.result, assignments)

                                    sentence_copy.pattern = self.apply_substitution(sentence_copy.pattern,
                                                                                    [*substitutions[j],
                                                                                     *self.format_function,
                                                                                     *self.default_format_function])
                                    sentence_copy.pattern = self.apply_assignment(sentence_copy.pattern,
                                                                                  assignments)

                                    sentence_copy.result = self.apply_substitution(sentence_copy.result,
                                                                                   [*substitutions[j],
                                                                                    *self.format_function,
                                                                                    *self.default_format_function])
                                    sentence_copy.result = self.apply_assignment(sentence_copy.result,
                                                                                 assignments)
                                    if sentence_copy not in sentence_result:
                                        sentence_result.append(sentence_copy)

                                    j += 1

                        function.sentences.extend(sentence_result)

                self.system_condition = []
                for function in ast.functions:
                    if isinstance(function, Definition):
                        for sentence in function.sentences:
                            for condition in sentence.conditions:
                                if not has_repeated_variable(condition.result):
                                    eq = Equation(condition.result, condition.pattern, EqType.Expr,
                                                  function.name, function.sentences.index(sentence),
                                                  sentence)
                                    self.system_condition.append(eq)

                substitutions_condition, assignments_condition = self.calculate_equation(
                    self.system_condition, ast)
                substitutions_variant = make_alternative(substitutions_condition)
                for substitution_var in substitutions_variant:
                    for subst in substitutions:
                        substitution_var.extend(subst)

                substitutions = substitutions_variant

                assignments.extend(assignments_condition)

            if len(self.system_result) == 0:
                for function in ast.functions:
                    if isinstance(function, Definition):
                        sentence_result = []
                        while len(function.sentences) > 0:
                            i = 0
                            new_sentence = deepcopy(function.sentences[0])
                            del function.sentences[0]
                            if len(substitutions) > 0:
                                while i < len(substitutions):

                                    sentence_first = deepcopy(new_sentence)

                                    sentence_first.pattern = self.apply_substitution(sentence_first.pattern,
                                                                                     [*substitutions[i],
                                                                                      *self.format_function,
                                                                                      *self.default_format_function])
                                    sentence_first.pattern = self.apply_assignment(sentence_first.pattern, assignments)

                                    sentence_first.result = self.apply_substitution(sentence_first.result,
                                                                                    [*substitutions[i],
                                                                                     *self.format_function,
                                                                                     *self.default_format_function])
                                    sentence_first.result = self.apply_assignment(sentence_first.result, assignments)

                                    if sentence_first not in sentence_result:
                                        sentence_result.append(sentence_first)
                                    i += 1
                            else:
                                if new_sentence not in sentence_result:
                                    sentence_result.append(new_sentence)

                        function.sentences.extend(sentence_result)

                format_substitution_function = get_substitution_function_format(ast)
                format_functions_result = self.generalization_expression(format_substitution_function)

                if self.is_fixed_point(self.format_function, format_functions_result):
                    break
                else:
                    self.format_function = format_functions_result

                    if DEBUG_MODE:
                        print(LINE_DELIMITER)
                        for assignment in assignments:
                            print(assignment)
                        for substitution_variant in substitutions:
                            print(', '.join(map(str, substitution_variant)))
                        for format_function in self.format_function:
                            print(format_function)
                        print(LINE_DELIMITER)
            else:
                for function in ast.functions:
                    if isinstance(function, Definition):
                        sentence_result = []
                        while len(function.sentences) > 0:
                            i = 0
                            new_sentence = deepcopy(function.sentences[0])
                            del function.sentences[0]
                            if len(substitutions) > 0:
                                while i < len(substitutions):

                                    sentence_first = deepcopy(new_sentence)

                                    sentence_first.pattern = self.apply_substitution(sentence_first.pattern,
                                                                                     [*substitutions[i],
                                                                                      *self.format_function,
                                                                                      *self.default_format_function])
                                    sentence_first.pattern = self.apply_assignment(sentence_first.pattern, assignments)

                                    sentence_first.result = self.apply_substitution(sentence_first.result,
                                                                                    [*substitutions[i],
                                                                                     *self.format_function,
                                                                                     *self.default_format_function])
                                    sentence_first.result = self.apply_assignment(sentence_first.result, assignments)

                                    if sentence_first not in sentence_result:
                                        sentence_result.append(sentence_first)
                                    i += 1
                            else:
                                if new_sentence not in sentence_result:
                                    sentence_result.append(new_sentence)

                        function.sentences.extend(sentence_result)

                for eq in self.system_result:
                    eq.sentence.no_substitution = False

                    # eq.left_part = self.apply_substitution(eq.left_part,
                    #                                        [*self.format_function, *self.default_format_function])
                    # eq.right_part = self.apply_substitution(eq.right_part,
                    #                                         [*self.format_function, *self.default_format_function])

                substitutions_result, assignments_result = self.calculate_equation(
                    self.system_result, ast)

                substitutions_variant = make_alternative(substitutions_result)
                for substitution_var in substitutions_variant:
                    for subst in substitutions:
                        substitution_var.extend(subst)

                assignments.extend(assignments_result)

                substitutions = substitutions_variant

                for function in ast.functions:
                    if isinstance(function, Definition):
                        sentence_result = []
                        while len(function.sentences) > 0:
                            i = 0
                            if len(substitutions) > 0:
                                sentence_copy = deepcopy(function.sentences[0])
                                del function.sentences[0]
                                while i < len(substitutions):
                                    new_sentence = deepcopy(sentence_copy)

                                    if sentence_copy.has_call:

                                        new_sentence.pattern = self.apply_substitution(new_sentence.pattern,
                                                                                       [*substitutions[i],
                                                                                        *self.format_function,
                                                                                        *self.default_format_function])
                                        new_sentence.pattern = self.apply_assignment(new_sentence.pattern, assignments)

                                        new_sentence.result = self.apply_substitution(new_sentence.result,
                                                                                      [*substitutions[i],
                                                                                       *self.format_function,
                                                                                       *self.default_format_function])
                                        new_sentence.result = self.apply_assignment(new_sentence.result, assignments)

                                        if new_sentence not in sentence_result:
                                            sentence_result.append(new_sentence)
                                    else:
                                        if new_sentence not in sentence_result:
                                            sentence_result.append(new_sentence)

                                    i += 1

                        function.sentences.extend(sentence_result)

                format_substitution_function = get_substitution_function_format(ast)
                format_functions_result = self.generalization_expression(format_substitution_function)

                if self.is_fixed_point(self.format_function, format_functions_result):
                    break
                else:
                    self.format_function = format_functions_result

                    if DEBUG_MODE:
                        print(LINE_DELIMITER)
                        for assignment in assignments:
                            print(assignment)
                        for substitution in substitutions:
                            print(', '.join(map(str, substitution)))
                        for format_function in self.format_function:
                            print(format_function)
                        print(LINE_DELIMITER)

    def is_fixed_point(self, before_format, after_format):
        common_format = dict()
        for format_function in before_format:
            if format_function.left_part not in common_format:
                common_format[format_function.left_part] = [format_function.right_part]
            else:
                common_format[format_function.left_part].append(format_function.right_part)

        for format_function in after_format:
            if format_function.left_part not in common_format:
                common_format[format_function.left_part] = [format_function.right_part]
            else:
                common_format[format_function.left_part].append(format_function.right_part)

        for key_group in common_format.keys():
            if len(common_format[key_group]) == 1 or len(common_format[key_group][0].terms) != len(
                    common_format[key_group][1].terms):
                return False
            else:
                for i in range(len(common_format[key_group][0].terms)):

                    if isinstance(common_format[key_group][0].terms[i], StructuralBrackets) and \
                            isinstance(common_format[key_group][1].terms[i], StructuralBrackets):

                        before_common_format_brackets = [
                            Substitution(key_group, Expression(common_format[key_group][0].terms[i].value))]
                        after_common_format_brackets = [
                            Substitution(key_group, Expression(common_format[key_group][1].terms[i].value))]

                        if not self.is_fixed_point(before_common_format_brackets, after_common_format_brackets):
                            return False

                    elif not (isinstance(common_format[key_group][0].terms[i], Variable) and
                              isinstance(common_format[key_group][1].terms[i], Variable) and
                              common_format[key_group][0].terms[i].type_variable == common_format[key_group][1].terms[
                                  i].type_variable) and \
                            not common_format[key_group][0].terms[i] == common_format[key_group][1].terms[i]:
                        return False

        return True

    def match_equation(self, equation, substitution, assignments, system):
        eq = deepcopy(equation)
        if eq.left_part.terms == [] and eq.right_part.terms == []:
            return "Success", substitution, assignments, system, eq
        else:
            if eq.type_equation == EqType.Term:
                term_left = eq.left_part.terms[0]
                term_right = eq.right_part.terms[0]

                if isinstance(term_right, Variable) and isinstance(term_left, Variable) and term_left == term_right:
                    return "Success", substitution, assignments, system, eq
                if is_symbol(term_left) and is_symbol(term_right) and term_left == term_right:
                    return "Success", substitution, assignments, system, eq
                elif (isinstance(term_right, Variable) and term_right.type_variable == Type.t) and \
                        ((isinstance(term_left, Variable) and
                          (term_left.type_variable == Type.s or term_left.type_variable == Type.t))
                         or isinstance(term_left, StructuralBrackets) or is_symbol(term_left)):
                    assignments.append(Assignment(Expression([term_left]), Expression([term_right])))
                    return "Success", substitution, assignments, system, eq
                elif ((isinstance(term_right, Variable) and term_right.type_variable == Type.s) or is_symbol(
                        term_right)) \
                        and isinstance(term_left, Variable) and term_left.type_variable == Type.t:
                    substitution.append(Substitution(Expression([term_left]), Expression([term_right])))
                    return "Success", substitution, assignments, system, eq
                elif (isinstance(term_right, Variable) and term_right.type_variable == Type.s) and \
                        ((isinstance(term_left, Variable) and term_left.type_variable == Type.s) or is_symbol(
                            term_left)):
                    assignments.append(Assignment(Expression([term_left]), Expression([term_right])))
                    return "Success", substitution, assignments, system, eq
                elif (isinstance(term_left, Variable) and term_left.type_variable == Type.s) and is_symbol(
                        term_right):
                    if term_left.index > 0:
                        substitution.append(Substitution(Expression([term_left]), Expression([term_right])))
                    return "Success", substitution, assignments, system, eq
                elif isinstance(term_left, StructuralBrackets) and isinstance(term_right, StructuralBrackets):
                    eq.type_equation = EqType.Expr
                    eq.left_part.terms = Expression(term_left.value).terms
                    eq.right_part.terms = Expression(term_right.value).terms
                    return self.match_equation(eq, substitution, assignments, system)
                elif isinstance(term_right, StructuralBrackets) and isinstance(term_left, Variable) \
                        and term_left.type_variable == Type.t:
                    index = generate_index()
                    e_variable = Variable("generated%d".format(index), Type.e, None, index)
                    substitution.append(Substitution(Expression([term_left]),
                                                     Expression([StructuralBrackets([e_variable])])))

                    eq.type_equation = EqType.Expr
                    eq.left_part.terms = Expression([e_variable]).terms
                    eq.right_part.terms = Expression(term_right.value).terms
                    return self.match_equation(eq, substitution, assignments, system)
                else:
                    return "Failure", [], [], system, eq

            else:
                if len(eq.left_part.terms) > 0 and is_hard_term(eq.left_part.terms[0]) \
                        and len(eq.right_part.terms) > 0 and is_hard_term(eq.right_part.terms[0]):

                    subst = deepcopy(substitution)
                    assign = deepcopy(assignments)
                    syst = deepcopy(system)

                    status, subst, assign, syst, e = self.match_equation(
                        Equation(Expression([eq.left_part.terms[0]]),
                                 Expression([eq.right_part.terms[0]]),
                                 EqType.Term, eq.func_name,
                                 eq.index_sentence, eq.sentence), subst,
                        assign, syst)

                    eq.left_part.terms = eq.left_part.terms[1:]
                    eq.right_part.terms = eq.right_part.terms[1:]

                    if status == "Success":
                        substitution = subst
                        system = syst
                        assignments = assign
                        return self.match_equation(eq, substitution, assignments, system)
                    else:
                        return "Failure", substitution, assignments, system, eq

                elif len(eq.left_part.terms) > 0 and is_hard_term(eq.left_part.terms[-1]) \
                        and len(eq.right_part.terms) > 0 and is_hard_term(eq.right_part.terms[-1]):

                    subst = deepcopy(substitution)
                    assign = deepcopy(assignments)
                    syst = deepcopy(system)

                    status, subst, assign, syst, e = self.match_equation(
                        Equation(Expression([eq.left_part.terms[-1]]),
                                 Expression([eq.right_part.terms[-1]]),
                                 EqType.Term, eq.func_name,
                                 eq.index_sentence, eq.sentence), subst,
                        assign, syst)

                    eq.left_part.terms = eq.left_part.terms[:-1]
                    eq.right_part.terms = eq.right_part.terms[:-1]

                    if status == "Success":
                        substitution = subst
                        assignments = assign
                        system = syst
                        return self.match_equation(eq, substitution, assignments, system)
                    else:
                        return "Failure", substitution, assignments, system, eq

                if len(eq.left_part.terms) > 0 and isinstance(eq.left_part.terms[0], Variable) and \
                        eq.left_part.terms[0].type_variable == Type.e and len(eq.right_part.terms) > 0 \
                        and is_hard_term(eq.right_part.terms[0]):
                    index = generate_index()
                    t_variable = Variable("generated%d".format(index), Type.t, None, index)
                    index = generate_index()
                    e_variable = Variable("generated%d".format(index), Type.e, None, index)
                    substitution_result = [*substitution,
                                           Substitution(Expression([eq.left_part.terms[0]]), Expression([]),
                                                        Expression([t_variable, e_variable]))]

                    return "Success", substitution_result, assignments, system, eq

                if len(eq.left_part.terms) > 0 and isinstance(eq.left_part.terms[-1], Variable) and \
                        eq.left_part.terms[-1].type_variable == Type.e and len(eq.right_part.terms) > 0 \
                        and is_hard_term(eq.right_part.terms[-1]):
                    index = generate_index()
                    e_variable = Variable("generated%d".format(index), Type.e, None, index)
                    index = generate_index()
                    t_variable = Variable("generated%d".format(index), Type.t, None, index)
                    substitution_result = [*substitution,
                                           Substitution(Expression([eq.left_part.terms[-1]]), Expression([]),
                                                        Expression([e_variable, t_variable]))]

                    return "Success", substitution_result, assignments, system, eq

                if len(eq.left_part.terms) == 0:

                    if len(eq.right_part.terms) > 0 and is_hard_term(eq.right_part.terms[0]):
                        return "Failure", [], [], system, eq

                    if len(eq.right_part.terms) == 0:
                        return "Success", substitution, assignments, system, eq

                    if all(isinstance(term, Variable) and term.type_variable == Type.e for term in
                           eq.right_part.terms):
                        for term in eq.left_part.terms:
                            assignments.append(Assignment(Expression([]), Expression([term])))
                        return "Success", substitution, assignments, system, eq

                    return "Undefined", [], [], system, eq

                if len(eq.right_part.terms) == 0:
                    if all(isinstance(term, Variable) and term.type_variable == Type.e for term in
                           eq.left_part.terms):
                        for term in eq.left_part.terms:
                            substitution.append(Substitution(Expression([term]), Expression([])))
                        return "Success", substitution, assignments, system, eq
                    else:
                        return "Failure", [], [], system, eq

                if len(eq.right_part.terms) == 1:
                    if isinstance(eq.right_part.terms[0], Variable) and eq.right_part.terms[0].type_variable == Type.e:
                        assignments.append(
                            Assignment(Expression(eq.left_part.terms), Expression([eq.right_part.terms[0]])))
                        return "Success", substitution, assignments, system, eq
                    else:
                        return "Failure", [], [], system, eq

                else:
                    return "Undefined", [], [], system, eq

    def generalization_term_rec(self, term_left, term_right):
        if is_symbol(term_left) and is_symbol(term_right):
            if term_left == term_right:
                return term_left
            else:
                index = generate_index()
                return Variable("generated%d".format(index), Type.s, None, index)
        if (isinstance(term_left, Variable) and term_left.type_variable == Type.s) and is_symbol(term_right):
            return term_left
        if (isinstance(term_right, Variable) and term_right.type_variable == Type.s) and is_symbol(term_left):
            return term_right
        if (isinstance(term_right, Variable) and term_right.type_variable == Type.s) and \
                (isinstance(term_left, Variable) and term_left.type_variable == Type.s):
            if term_left == term_right:
                return term_left
            else:
                index = generate_index()
                return Variable("generated%d".format(index), Type.s, None, index)
        if isinstance(term_left, Variable) and term_left.type_variable == Type.t:
            return term_left
        if isinstance(term_right, Variable) and term_right.type_variable == Type.t:
            return term_right
        if isinstance(term_left, StructuralBrackets) and isinstance(term_right, StructuralBrackets):
            result_terms = self.generalization([Expression(term_left.value), Expression(term_right.value)]).terms
            return StructuralBrackets(result_terms)
        if isinstance(term_left, StructuralBrackets):
            index = generate_index()
            return Variable("generated%d".format(index), Type.t, None, index)
        if isinstance(term_right, StructuralBrackets):
            index = generate_index()
            return Variable("generated%d".format(index), Type.t, None, index)

    def generalization_term(self, terms):
        terms_copy = deepcopy(terms)
        if len(terms_copy) > 1:
            while len(terms_copy) > 1:
                term = self.generalization_term_rec(terms_copy.pop(), terms_copy.pop())
                terms_copy.append(term)
            return terms_copy[0]
        else:
            term = terms_copy.pop()
            result_term = self.generalization_term_rec(term, deepcopy(term))
            return result_term

    def generalization(self, patterns):
        if len(patterns) > 1:
            for pattern in patterns:
                special_value = self.get_none(pattern)
                if special_value is not None:
                    patterns.remove(pattern)
                    return self.generalization(patterns)
        elif len(patterns) == 1:
            special_value = self.get_none(patterns[0])
            if special_value is not None:
                return Expression([special_value])

        all_empty = all(len(pattern.terms) == 0 for pattern in patterns)
        any_empty = any(len(pattern.terms) == 0 for pattern in patterns)

        if all_empty:
            return Expression([])
        elif any_empty:
            index = generate_index()
            return Expression([Variable("generated%d".format(index), Type.e, None, index)])
        else:
            left = [pattern.terms[0] for pattern in patterns]
            right = [pattern.terms[-1] for pattern in patterns]

            left_e = any(isinstance(term, Variable) and term.type_variable == Type.e for term in left)
            right_e = any(isinstance(term, Variable) and term.type_variable == Type.e for term in right)

            if left_e:
                if right_e:
                    if left == right and len(left) == 1:
                        return Expression(left)
                    else:
                        index = generate_index()
                        return Expression([Variable("generated%d".format(index), Type.e, None, index)])
                else:
                    for pattern in patterns:
                        pattern.terms.pop()
                    return Expression([*self.generalization(patterns).terms, self.generalization_term(right)])
            else:
                if right_e:
                    for pattern in patterns:
                        pattern.terms.pop(0)
                    return Expression([self.generalization_term(left), *self.generalization(patterns).terms])
                else:
                    left_t = [self.generalization_term(left)]
                right_t = [self.generalization_term(right)]
                if calc_complexity(left_t) >= calc_complexity(right_t):
                    for pattern in patterns:
                        pattern.terms.pop(0)
                    return Expression([*left_t, *self.generalization(patterns).terms])
                else:
                    for pattern in patterns:
                        pattern.terms.pop()
                    return Expression([*self.generalization(patterns).terms, *right_t])

    def print_format_function(self):
        for error_message in self.error_messages:
            sys.stderr.write(error_message)

        result_str = "/* result program refalcheck */\n\n"

        for format_function in self.format_function:
            format_function.right_part.terms = append_char_term(format_function.right_part.terms)

        for function in self.ast.functions:
            if isinstance(function, Definition):
                in_format = Expression([SpecialVariable(function.name, SpecialType.in_function)])
                out_format = Expression([SpecialVariable(function.name, SpecialType.out_function)])
                if function.is_entry:
                    result_str = result_str + function.name + " "
                else:
                    result_str = result_str + "*" + function.name + " "
                for format_function in self.format_function:
                    if format_function.left_part == in_format:
                        pattern = format_function.right_part
                        result_str += (' '.join(list(map(self.print_term, pattern.terms))) + ' ')
                    if format_function.left_part == out_format:
                        result_str += ("=" + ' ')
                        result = format_function.right_part
                        result_str += (' '.join(list(map(self.print_term, result.terms))) + '')
                result_str += "\n"

        return result_str

    def print_expr(self, expr):
        return ' '.join(list(map(self.print_term, expr.terms)))

    def print_equation(self, equation):
        result_str = ""
        result_str += (' '.join(list(map(self.print_term, equation.left_part.terms))))
        result_str += " : "
        result_str += (' '.join(list(map(self.print_term, equation.right_part.terms))))
        return result_str

    def print_term(self, term):
        if isinstance(term, StructuralBrackets):
            output = '('
            output += ' '.join(list(map(self.print_term, term.value)))
            output += ')'
            return output
        if isinstance(term, Variable):
            return term.type_variable.name
        else:
            return term.__str__()



#! python -v
# -*- coding: utf-8 -*-

from refal.position import *
from refal.tokens import *
from refal.constants import *

from copy import deepcopy
import sys


def next_or_current(obj):
    return next(deepcopy(obj), obj)


def refactor_char_token(tokens):
    i = 0
    while i < len(tokens):
        if isinstance(tokens[i], CharacterToken):
            if len(tokens[i].value) > 3:
                new_tokens = []
                ch = ""
                position = deepcopy(tokens[i].coords)
                for j in range(len(tokens[i].value)):
                    if tokens[i].value[j] == "'" and ch == "":
                        continue
                    if ch == "\\" and tokens[i].value[j] != "\\":
                        ch += tokens[i].value[j]
                        ch = "'" + ch + "'"
                        new_tokens.append(CharacterToken(ch, position))
                        ch = ""
                    elif tokens[i].value[j] == "\\":
                        ch += tokens[i].value[j]
                    else:
                        ch += tokens[i].value[j]
                        ch = "'" + ch + "'"
                        new_tokens.append(CharacterToken(ch, position))
                        ch = ""
                if len(new_tokens) > 1:
                    tokens = tokens[:i] + new_tokens + tokens[i+1:]
                    i = -1
        i += 1
    return tokens


class Lexer(object):

    def __init__(self, program):
        self.program = program
        self.cur = Position(program)

    def read_keyword(self):
        new_pos = next_or_current(self.cur)
        s = "$"
        if new_pos.letter() == "E":
            s += new_pos.letter()
            new_pos = next_or_current(new_pos)
            if new_pos.letter() == "N" or new_pos.letter() == "X":
                s += new_pos.letter()
                new_pos = next_or_current(new_pos)
                if new_pos.letter() == "T":
                    s += new_pos.letter()
                    new_pos = next_or_current(new_pos)
                    if new_pos.letter() == "E" and s[-2] != "N" or new_pos.letter() == "R":
                        s += new_pos.letter()
                        new_pos = next_or_current(new_pos)
                        if new_pos.letter() == "N" and s[-1] != "E" or new_pos.letter() == "R" \
                                or new_pos.letter() == "Y" and s[2] == "N":
                            s += new_pos.letter()
                            if s == "$EXTRN" or s == "$ENTRY":
                                return KeywordToken(s, Fragment(self.cur, new_pos))
                            new_pos = next_or_current(new_pos)
                            if new_pos.letter() == "N":
                                s += new_pos.letter()
                                new_pos = next_or_current(new_pos)
                                if new_pos.letter() == "A":
                                    s += new_pos.letter()
                                    new_pos = next_or_current(new_pos)
                                    if new_pos.letter() == "L":
                                        s += new_pos.letter()
                                        new_pos = next_or_current(new_pos)
                                        if not new_pos.is_letter():
                                            return KeywordToken(s, Fragment(self.cur, new_pos))
                                elif new_pos != next_or_current(new_pos) and not new_pos.is_letter():
                                    return KeywordToken(s, Fragment(self.cur, new_pos))
        else:
            return UnknownToken("$", Fragment(self.cur, self.cur))
        return UnknownToken("$", Fragment(self.cur, self.cur))

    def read_macro_number(self):
        new_pos = next_or_current(self.cur)
        s = self.cur.letter()
        if new_pos != next_or_current(new_pos) and (new_pos.is_decimal_digit()):
            s += new_pos.letter()
            while new_pos != next_or_current(new_pos) and next_or_current(new_pos).is_decimal_digit():
                new_pos = next_or_current(new_pos)
                s += new_pos.letter()
            return NumberToken(int(s), Fragment(self.cur, new_pos))

        return NumberToken(int(s), Fragment(self.cur, self.cur))

    def read_ident_or_variable(self):
        new_pos = next_or_current(self.cur)
        s = self.cur.letter()
        if new_pos.letter() == "." and s in ["s", "t", "e"]:
            s += new_pos.letter()
            new_pos = next_or_current(new_pos)
            if new_pos.is_decimal_digit():
                s += new_pos.letter()
                while new_pos != next_or_current(new_pos) and next_or_current(new_pos).is_decimal_digit():
                    new_pos = next_or_current(new_pos)
                    s += new_pos.letter()
                return VariableToken(s, Fragment(self.cur, new_pos))
            elif new_pos.is_latin_letter():
                s += new_pos.letter()
                while new_pos != next_or_current(new_pos) and (next_or_current(new_pos).is_letter_or_digit() or
                                                               next_or_current(new_pos).letter() in ["-", "_"]):
                    new_pos = next_or_current(new_pos)
                    s += new_pos.letter()
                return VariableToken(s, Fragment(self.cur, new_pos))
            else:
                return UnknownToken(s, Fragment(self.cur, new_pos))
        elif new_pos.letter() == ".":
            return IdentToken(s, Fragment(self.cur, next(self.cur, self.cur)))
        if new_pos != next_or_current(new_pos) and (new_pos.is_letter_or_digit() or new_pos.letter() in ["-", "_"]):
            s += new_pos.letter()
            # new_pos = next(copy.deepcopy(new_pos),  new_pos)
            while new_pos != next_or_current(new_pos) and (next_or_current(new_pos).is_letter_or_digit() or
                                                           next_or_current(new_pos).letter() in ["-", "_"]):
                new_pos = next_or_current(new_pos)
                s += new_pos.letter()
            return IdentToken(s, Fragment(self.cur, new_pos))
        return IdentToken(s, Fragment(self.cur, new_pos))

    def read_symbol(self):
        new_pos = next_or_current(self.cur)
        s = self.cur.letter()
        while new_pos != next_or_current(new_pos) and new_pos.letter() != "\"":
            if new_pos.letter() == "\\":
                s += new_pos.letter()
                new_pos = next_or_current(new_pos)
                if new_pos.letter() in ["\\", "n", "r", "t", "\"", "\'", "(", ")", "<", ">"]:
                    s += new_pos.letter()
                    new_pos = next_or_current(new_pos)
                elif new_pos.letter() == "x":
                    s += new_pos.letter()
                    new_pos = next_or_current(new_pos)
                    if new_pos != next_or_current(new_pos):
                        for _ in range(0, 2):
                            if new_pos.is_decimal_digit() or ("a" <= new_pos.letter() <= "f" or
                                                              "A" <= new_pos.letter() <= "F"):
                                s += new_pos.letter()
                                new_pos = next_or_current(new_pos)
                            else:
                                return UnknownToken(s, Fragment(self.cur, new_pos))
                        return CompositeSymbolToken(s, Fragment(self.cur, new_pos))
                else:
                    return UnknownToken(s, Fragment(self.cur, new_pos))
            else:
                s += new_pos.letter()
                new_pos = next_or_current(new_pos)
        if new_pos != next_or_current(new_pos):
            s += new_pos.letter()
            return CompositeSymbolToken(s, Fragment(self.cur, new_pos))
        else:
            return UnknownToken(s, Fragment(self.cur, new_pos))

    def read_chars(self):
        new_pos = next_or_current(self.cur)
        s = self.cur.letter()
        while new_pos != next_or_current(new_pos) and new_pos.letter() != "\'":
            if new_pos.letter() == "\\":
                s += new_pos.letter()
                new_pos = next_or_current(new_pos)
                if new_pos.letter() in ["\\", "n", "r", "t", "\"", "\'", "(", ")", "<", ">"]:
                    s += new_pos.letter()
                    new_pos = next_or_current(new_pos)
                elif new_pos.letter() == "x":
                    s += new_pos.letter()
                    new_pos = next_or_current(new_pos)
                    if new_pos != next_or_current(new_pos):
                        for _ in range(0, 2):
                            if new_pos.is_decimal_digit() or ("a" <= new_pos.letter() <= "f" or
                                                              "A" <= new_pos.letter() <= "F"):
                                s += new_pos.letter()
                                new_pos = next_or_current(new_pos)
                            else:
                                return UnknownToken(s, Fragment(self.cur, new_pos))
                        return CharacterToken(s, Fragment(self.cur, new_pos))
                else:
                    return UnknownToken(s, Fragment(self.cur, new_pos))
            else:
                s += new_pos.letter()
                new_pos = next_or_current(new_pos)
        if new_pos != next_or_current(new_pos):
            s += new_pos.letter()
            return CharacterToken(s, Fragment(self.cur, new_pos))
        else:
            return UnknownToken(s, Fragment(self.cur, new_pos))

    def read_left_call_or_mark(self):
        new_pos = next_or_current(self.cur)
        s = self.cur.letter()
        if new_pos != next_or_current(new_pos) and new_pos.letter() in ["+", "-", "*", "/", "%", "?"]:
            s += new_pos.letter()
            return LeftBracketToken(s, Fragment(self.cur, new_pos))
        while new_pos.is_white_space():
            new_pos = next_or_current(new_pos)
        if new_pos != next_or_current(new_pos):
            s += new_pos.letter()
            while new_pos != next_or_current(new_pos) and (next_or_current(new_pos).is_letter_or_digit() or
                                                           next_or_current(new_pos).letter() in ["-", "_"]):
                new_pos = next_or_current(new_pos)
                s += new_pos.letter()
            return LeftBracketToken(s, Fragment(self.cur, new_pos))
        else:
            return UnknownToken(s, Fragment(self.cur, new_pos))

    def read_many_line_comment(self):
        s = self.cur.letter()
        new_pos = next_or_current(self.cur)
        s += new_pos.letter()
        new_pos = next(new_pos, new_pos)
        while new_pos != next_or_current(new_pos) and not (new_pos.letter() == "*" and
                                                           next_or_current(new_pos).letter() == "/"):
            s += new_pos.letter()
            new_pos = next_or_current(new_pos)
        s += new_pos.letter()
        new_pos = next(new_pos, new_pos)
        s += new_pos.letter()
        if DEBUG_MODE:
            sys.stdout.write("Recognize many-line comment[" + str(Fragment(self.cur, new_pos)) + "]: " + s + "\n")
        new_pos = next_or_current(new_pos)
        self.cur = new_pos

    def read_comment(self):
        new_pos = next_or_current(self.cur)
        s = "*"
        while new_pos != next_or_current(new_pos) and not (new_pos.letter() == "\n"):
            s += new_pos.letter()
            new_pos = next_or_current(new_pos)
        new_pos = next_or_current(new_pos)
        if DEBUG_MODE:
            sys.stdout.write("Recognize one-line comment[" + str(Fragment(self.cur, new_pos)) + "]: " + s + "\n")
        self.cur = new_pos

    def next_token(self):
        tok = UnknownToken(self.cur.letter(), Fragment(self.cur, self.cur))
        while tok.tag == DomainTag.Unknown:
            while self.cur.is_white_space():
                self.cur = next(self.cur, self.cur)
            if self.cur.letter() == "/" and self.cur != next_or_current(self.cur) and \
                    next_or_current(self.cur).letter() == "*":
                self.read_many_line_comment()
                continue
            if self.cur.letter() == "*":
                self.read_comment()
                continue
            if self.cur.is_eof():
                tok = EopToken(Fragment(self.cur, self.cur))
            else:
                if self.cur.letter() == "$":
                    read_tok = self.read_keyword()
                elif self.cur.letter() == "\"":
                    read_tok = self.read_symbol()
                elif self.cur.letter() == "\'":
                    read_tok = self.read_chars()
                elif self.cur.letter() == "<":
                    read_tok = self.read_left_call_or_mark()
                elif self.cur.letter() in ['[', '(', ')', '>', '=', ';', ':', ',', '{', '}', ']']:
                    read_tok = MarkSignToken(self.cur.letter(), Fragment(self.cur, self.cur))
                elif self.cur.is_decimal_digit():
                    read_tok = self.read_macro_number()
                elif self.cur.is_latin_letter():
                    read_tok = self.read_ident_or_variable()
                else:
                    read_tok = UnknownToken(self.cur.letter(), Fragment(self.cur, self.cur))
                if read_tok.tag == DomainTag.Unknown:
                    sys.stderr.write("Token[" + str(read_tok) + "]: unrecognized token\n")
                    self.cur = next(self.cur, self.cur)
                else:
                    self.cur = read_tok.coords.following
                    self.cur = next_or_current(self.cur)
                    tok = read_tok
        return tok



#! python -v
# -*- coding: utf-8 -*-

from refal.tokens import *
from refal.ast import *

import sys


class ParserRefal(object):

    def __init__(self, tokens):
        self.tokens = tokens
        self.iteratorTokens = iter(self.tokens)
        self.cur_token = next(self.iteratorTokens)
        self.isError = False
        self.ast = None

    def get_variables(self, term):
        if isinstance(term, StructuralBrackets):
            variable = []
            for term in term.value:
                variable.extend(self.get_variables(term))
            return variable
        elif isinstance(term, CallBrackets):
            variable = []
            for term in term.content:
                variable.extend(self.get_variables(term))
            return variable
        elif isinstance(term, Variable):
            return [term.value]
        else:
            return []

    def check_call(self, term, ast_type):
        for term_content in term.content:
            if isinstance(term_content, CallBrackets):
                self.check_call(term_content, ast_type)

        has_body = False
        has_definition_type = False

        for function_new in self.ast.functions:
            if isinstance(function_new, Definition):
                if term.value == function_new.name:
                    has_body = True
                    break

        if not has_body:
            for function_definition_type in ast_type.functions:
                if isinstance(function_definition_type, DefinitionType):
                    if term.value == function_definition_type.name:
                        has_definition_type = True

        if not has_definition_type and not has_body:
            sys.stderr.write("Error. Unknown function %s\n" % term.value)
            self.isError = True

    def semantics_call(self, ast_type):
        if not self.isError:
            for function in self.ast.functions:
                if isinstance(function, Definition):
                    for sentence in function.sentences:
                        for term in sentence.result.terms:
                            if isinstance(term, CallBrackets):
                                self.check_call(term, ast_type)

    def semantics_variable(self, sentences):
        for sentence in sentences:
            variables = []
            for term in sentence.pattern.terms:
                variables.extend(self.get_variables(term))
            for condition in sentence.conditions:
                for term_condition in condition.pattern.terms:
                    variables.extend(self.get_variables(term_condition))
                for term_result in condition.result.terms:
                    out_variables = self.get_variables(term_result)
                    for variable in out_variables:
                        if variable not in variables:
                            sys.stderr.write(
                                "Error. Variable %s isn't found in previous sentence" % variable)
                            self.isError = True
            if sentence.block:
                self.semantics_variable(sentence.block)
            for term in sentence.result.terms:
                out_variables = self.get_variables(term)
                for variable in out_variables:
                    if variable not in variables:
                        sys.stderr.write("Error. Variable %s isn't found in previous sentence\n" % variable)
                        self.isError = True

    def semantics(self):
        names = set()
        for function in self.ast.functions:
            if function.name in names:
                sys.stderr.write("Error. Function %s already defined\n" % function.name)
                self.isError = True
            else:
                names.add(function.name)

        if not self.isError:
            for function in self.ast.functions:
                if isinstance(function, Definition):
                    for sentence in function.sentences:
                        for term in sentence.result.terms:
                            if isinstance(term, CallBrackets):
                                if term.value not in names:
                                    sys.stderr.write("Error. Function %s isn't defined\n" % term.value)
                                    self.isError = True
        if not self.isError:
            for function in self.ast.functions:
                if isinstance(function, Definition):
                    self.semantics_variable(function.sentences)

    def parse(self):
        self.ast = AST(self.parse_program())
        if self.cur_token.tag != DomainTag.Eop:
            self.isError = True
            sys.stderr.write("Error. Expected Token \"End_Of_Program\"\n")
        else:
            sys.stdout.write("Ok. Program satisfy grammar\n")

    # Program ::= Global*
    def parse_program(self):
        functions = self.parse_global()
        while self.cur_token.tag == DomainTag.Keyword or self.cur_token.tag == DomainTag.Ident or \
                (self.cur_token.tag == DomainTag.Mark_sign and self.cur_token.value == ";"):
            functions.extend(self.parse_global())
        return functions

    # Global ::= Externs | Function | ';';
    def parse_global(self):
        if self.cur_token.tag == DomainTag.Keyword and self.cur_token.value == "$ENTRY" or \
                self.cur_token.tag == DomainTag.Ident:
            function = self.parse_function()
            return function
        elif self.cur_token.tag == DomainTag.Keyword:
            externs = self.parse_externs()
            return externs
        elif self.cur_token.tag == DomainTag.Mark_sign and self.cur_token.value == ";":
            self.cur_token = next(self.iteratorTokens)
            return []

    # Externs ::= ExternKeyword 'Name' (',' 'Name')* ';';
    # ExternKeyword ::= '$EXTERN' | '$EXTRN' | '$EXTERNAL';
    def parse_externs(self):
        if self.cur_token.value == "$ENTRY":
            sys.stderr.write("Expected keyword: $EXTERN|$EXTRN|$EXTERNAL\n")
            self.isError = True
        else:
            self.cur_token = next(self.iteratorTokens)
            if self.cur_token.tag == DomainTag.Ident:
                externs = [Extern(self.cur_token.value, self.cur_token.coords)]
                self.cur_token = next(self.iteratorTokens)
                while self.cur_token.tag == DomainTag.Mark_sign and self.cur_token.value == ",":
                    self.cur_token = next(self.iteratorTokens)
                    if self.cur_token.tag != DomainTag.Ident:
                        sys.stderr.write("Expected name of external function\n")
                        self.isError = True
                        return []
                    else:
                        externs.append(Extern(self.cur_token.value, self.cur_token.coords))
                    self.cur_token = next(self.iteratorTokens)
                if not (self.cur_token.tag == DomainTag.Mark_sign and self.cur_token.value == ";"):
                    sys.stderr.write("Expected \";\" after $EXTERNAL\n")
                    self.isError = True
                else:
                    self.cur_token = next(self.iteratorTokens)
                    return externs
            else:
                sys.stderr.write("Expected name of external function\n")
                self.isError = True
        return []

    # Function ::= ('$ENTRY')? 'Name' Body;
    def parse_function(self):
        is_entry = False
        if self.cur_token.tag == DomainTag.Keyword and self.cur_token.value == "$ENTRY":
            self.cur_token = next(self.iteratorTokens)
            is_entry = True
        if self.cur_token.tag != DomainTag.Ident:
            sys.stderr.write("Expected name of function\n")
            self.isError = True
            return []
        else:
            function_name = self.cur_token.value
            position = self.cur_token.coords
            self.cur_token = next(self.iteratorTokens)
            body = self.parse_body()
            return [Definition(function_name, position, is_entry, body)]

    # Body ::= '{' Sentences '}';
    def parse_body(self):
        if self.cur_token.tag == DomainTag.Mark_sign and self.cur_token.value == "{":
            self.cur_token = next(self.iteratorTokens)
            sentences = self.parse_sentences()
            if self.cur_token.tag == DomainTag.Mark_sign and self.cur_token.value == "}":
                self.cur_token = next(self.iteratorTokens)
            else:
                sys.stderr.write("Expected \"}\" after declaring function\n")
                self.isError = True
            return sentences
        else:
            sys.stderr.write("Expected \"{\" after declaring function\n")
            self.isError = True
            return []

    # Sentences ::= Sentence (';' Sentences?)?;
    def parse_sentences(self):
        sentence = self.parse_sentence()
        if self.cur_token.tag == DomainTag.Mark_sign and self.cur_token.value == ";":
            self.cur_token = next(self.iteratorTokens)
            sentences = self.parse_sentences()
            sentence = [*sentence, *sentences]
            return sentence
        return sentence

    # Sentence ::= Pattern ( ('=' Result) | (',' Result ':' (Sentence | Body)) );
    def parse_sentence(self):
        pattern = self.parse_pattern()
        if self.cur_token.tag == DomainTag.Mark_sign and self.cur_token.value == "=":
            self.cur_token = next(self.iteratorTokens)
            result = self.parse_result()
            return [Sentence(pattern, [], result, [])]
        elif self.cur_token.tag == DomainTag.Mark_sign and self.cur_token.value == ",":
            self.cur_token = next(self.iteratorTokens)
            result = self.parse_result()
            if self.cur_token.tag == DomainTag.Mark_sign and self.cur_token.value == ":":
                self.cur_token = next(self.iteratorTokens)
                if self.cur_token.tag == DomainTag.Mark_sign and self.cur_token.value == "{":
                    body = self.parse_body()
                    return [Sentence(pattern, [], result, body)]
                else:
                    sentence = self.parse_sentence()[0]
                    return [
                        Sentence(pattern, [*sentence.conditions, Condition(result, sentence.pattern)], sentence.result,
                                 sentence.block)]
            else:
                sys.stderr.write("Expected \":\" after declaring result\n")
                self.isError = True
            return [Sentence(pattern, [], result, [])]
        return []
        # else:
        #     sys.stderr.write("Expected \"=\" or \",\" after declaring pattern\n")

    # Pattern ::= PatternTerm*;
    def parse_pattern(self):
        patternterm = self.parse_patternterm()
        while self.cur_token.tag == DomainTag.Ident or self.cur_token.tag == DomainTag.Number or \
                self.cur_token.tag == DomainTag.Characters or self.cur_token.tag == DomainTag.Composite_symbol \
                or self.cur_token.tag == DomainTag.Variable \
                or (self.cur_token.tag == DomainTag.Mark_sign and self.cur_token.value == "("):
            patternterm.extend(self.parse_patternterm())
        return Expression(patternterm)

    # PatternTerm ::= Common | '(' Pattern ')';
    def parse_patternterm(self):
        if self.cur_token.tag == DomainTag.Mark_sign and self.cur_token.value == "(":
            pattern = []
            self.cur_token = next(self.iteratorTokens)
            pattern.append(StructuralBrackets(self.parse_pattern().terms))
            if self.cur_token.tag == DomainTag.Mark_sign and self.cur_token.value == ")":
                self.cur_token = next(self.iteratorTokens)
                return pattern
            else:
                sys.stderr.write("Expected \")\" after declaring pattern\n")
                self.isError = True
            return []
        else:
            return self.parse_common()

    # Common ::= 'Name' | ''chars'' | ""common"" | '123' | 'e.Var' | ε;
    def parse_common(self):
        if self.cur_token.tag == DomainTag.Ident:
            token = self.cur_token
            self.cur_token = next(self.iteratorTokens)
            return [CompoundSymbol(token.value)]
        elif self.cur_token.tag == DomainTag.Characters:
            token = self.cur_token
            self.cur_token = next(self.iteratorTokens)
            return [Char(token.value)]
        elif self.cur_token.tag == DomainTag.Number:
            token = self.cur_token
            self.cur_token = next(self.iteratorTokens)
            return [Macrodigit(token.value)]
        elif self.cur_token.tag == DomainTag.Composite_symbol:
            token = self.cur_token
            self.cur_token = next(self.iteratorTokens)
            return [CompoundSymbol(token.value)]
        elif self.cur_token.tag == DomainTag.Variable:
            token = self.cur_token
            self.cur_token = next(self.iteratorTokens)
            return [Variable(token.value, Type[token.value[0]], token.coords)]
        else:
            return []

    # Result ::= ResultTerm*;
    def parse_result(self):
        resultterm = self.parse_result_term()
        while self.cur_token.tag == DomainTag.Ident or self.cur_token.tag == DomainTag.Number or \
                self.cur_token.tag == DomainTag.Characters or self.cur_token.tag == DomainTag.Composite_symbol \
                or self.cur_token.tag == DomainTag.Variable \
                or (self.cur_token.tag == DomainTag.Mark_sign and self.cur_token.value == "(") \
                or (self.cur_token.tag == DomainTag.Left_bracket):
            resultterm.extend(self.parse_result_term())
        return Expression(resultterm)

    # ResultTerm ::= Common | '(' Result ')' | '<Name' Result '>';
    def parse_result_term(self):
        if self.cur_token.tag == DomainTag.Mark_sign and self.cur_token.value == "(":
            self.cur_token = next(self.iteratorTokens)
            result = [StructuralBrackets(self.parse_result().terms)]
            if self.cur_token.tag == DomainTag.Mark_sign and self.cur_token.value == ")":
                self.cur_token = next(self.iteratorTokens)
                return result
            else:
                sys.stderr.write("Expected \")\" after declaring result\n")
                self.isError = True
            return []
        elif self.cur_token.tag == DomainTag.Left_bracket:
            func_name = self.cur_token.value[1:]
            pos = self.cur_token.coords
            self.cur_token = next(self.iteratorTokens)
            result = self.parse_result().terms
            if self.cur_token.tag == DomainTag.Mark_sign and self.cur_token.value == ">":
                self.cur_token = next(self.iteratorTokens)
            else:
                sys.stderr.write("Expected \">\" after declaring result\n")
                self.isError = True
            return [CallBrackets(func_name, pos, result)]
        else:
            return self.parse_common()


class ParserRefalType(object):

    def __init__(self, tokens):
        self.tokens = tokens
        self.iteratorTokens = iter(self.tokens)
        self.cur_token = next(self.iteratorTokens)
        self.isError = False
        self.ast = None

    def parse(self):
        self.ast = AST(self.parse_file(), True)
        if self.cur_token.tag != DomainTag.Eop:
            self.isError = True
            sys.stderr.write("Error. Expected Token \"End_Of_Program\"\n")
        else:
            sys.stdout.write("Ok. Program-Type satisfy grammar\n")

    def semantics(self):
        names = set()
        for function in self.ast.functions:
            if function.name in names:
                sys.stderr.write("Error. Function %s already defined\n" % function.name)
                self.isError = True
            else:
                names.add(function.name)

    # File ::= Function*
    def parse_file(self):
        function = self.parse_function()
        while self.cur_token.tag == DomainTag.Ident:
            function.extend(self.parse_function())
        return function

    # Function ::= 'Name' Format '=' Format ';'
    def parse_function(self):
        if self.cur_token.tag == DomainTag.Ident:
            func_name = self.cur_token.value
            pos = self.cur_token.coords
            self.cur_token = next(self.iteratorTokens)
            pattern = self.parse_format()
            result = []
            if self.cur_token.tag == DomainTag.Mark_sign and self.cur_token.value == "=":
                self.cur_token = next(self.iteratorTokens)
                result = self.parse_format()
            else:
                sys.stderr.write("Expected \"=\" after declaring name of function\n")
                self.isError = True
            if not (self.cur_token.tag == DomainTag.Mark_sign and self.cur_token.value == ";"):
                sys.stderr.write("Expected \";\" after function\n")
                self.isError = True
            else:
                self.cur_token = next(self.iteratorTokens)
                return [DefinitionType(func_name, pattern, result, pos)]
        return []

    # Format ::= Common ('e.Var' Common)?
    def parse_format(self):
        format_function = self.parse_common()
        if self.cur_token.tag == DomainTag.Variable and self.cur_token.value[0] == "e":
            format_function.append(Variable(self.cur_token.value, Type[self.cur_token.value[0]], self.cur_token.coords))
            self.cur_token = next(self.iteratorTokens)
            format_function.extend(self.parse_common())
        return Expression(format_function)

    # Common ::= ('Name' | ''chars'' | '123' | 's.Var' | 't.Var' | '(' Format ')')* | ε
    def parse_common(self):
        common_term = []
        while self.cur_token.tag == DomainTag.Ident or self.cur_token.tag == DomainTag.Number or \
                self.cur_token.tag == DomainTag.Characters or self.cur_token.tag == DomainTag.Composite_symbol \
                or (self.cur_token.tag == DomainTag.Variable and (self.cur_token.value[0] == "s" or
                                                                  self.cur_token.value[0] == "t")) \
                or (self.cur_token.tag == DomainTag.Mark_sign and self.cur_token.value == "("):
            if self.cur_token.tag == DomainTag.Mark_sign and self.cur_token.value == "(":
                self.cur_token = next(self.iteratorTokens)
                common_term.append(StructuralBrackets(self.parse_format().terms))
                if self.cur_token.tag == DomainTag.Mark_sign and self.cur_token.value == ")":
                    self.cur_token = next(self.iteratorTokens)
                else:
                    sys.stderr.write("Expected \")\" after declaring pattern\n")
                    self.isError = True
            else:
                if self.cur_token.tag == DomainTag.Ident:
                    token = self.cur_token
                    self.cur_token = next(self.iteratorTokens)
                    common_term.append(CompoundSymbol(token.value))
                elif self.cur_token.tag == DomainTag.Characters:
                    token = self.cur_token
                    self.cur_token = next(self.iteratorTokens)
                    common_term.append(Char(token.value))
                elif self.cur_token.tag == DomainTag.Number:
                    token = self.cur_token
                    self.cur_token = next(self.iteratorTokens)
                    common_term.append(Macrodigit(token.value))
                elif self.cur_token.tag == DomainTag.Composite_symbol:
                    token = self.cur_token
                    self.cur_token = next(self.iteratorTokens)
                    common_term.append(CompoundSymbol(token.value))
                elif self.cur_token.tag == DomainTag.Variable:
                    token = self.cur_token
                    self.cur_token = next(self.iteratorTokens)
                    common_term.append(Variable(token.value, Type[token.value[0]], token.coords))
        return common_term



#! python -v
# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod
from enum import Enum

from refal.constants import *


class AST(object):

    def __init__(self, functions, is_file_type=False):

        if not is_file_type:
            # Для доступности default-функций:
            default_functions = [Extern(name) for name in DEFAULT_FUNCTIONS]
            # default_functions = []
            self.functions = [*default_functions, *functions]
        else:
            self.functions = functions

    def __str__(self):
        return str("\n".join(list(map(str, self.functions))))


class Function(ABC):

    def __init__(self, name, pos):
        self.name = name
        self.pos = pos

    def __eq__(self, other):
        return isinstance(other, Function) and self.name == other.name

    @abstractmethod
    def __str__(self):
        return self.name


class Extern(Function):

    def __init__(self, name, pos=None):
        super(Extern, self).__init__(name, pos)

    def __eq__(self, other):
        return isinstance(other, Extern) and self.name == other.name

    def __str__(self):
        return "$EXTERN " + self.name


class Definition(Function):

    def __init__(self, name, pos, is_entry=False, sentences=None):
        super(Definition, self).__init__(name, pos)
        if sentences is None:
            sentences = []
        self.is_entry = is_entry
        self.sentences = sentences

    def __eq__(self, other):
        return isinstance(other, Definition) and self.name == other.name

    def __str__(self):
        return self.name + " {\n" + ";\n".join(list(map(str, self.sentences))) + ";\n}"


class DefinitionType(Function):

    def __init__(self, name, pattern, result, pos):
        super(DefinitionType, self).__init__(name, pos)
        self.pattern = pattern
        self.result = result

    def __eq__(self, other):
        return isinstance(other, DefinitionType) and self.name == other.name

    def __str__(self):
        return self.name + " " + str(self.pattern) + " = " + str(self.result)


class Sentence(object):

    def __init__(self, pattern, conditions, result, block):
        self.pattern = pattern
        self.conditions = conditions
        self.result = result
        self.block = block

        self.has_call = False
        self.no_substitution = False

    def __eq__(self, other):
        return isinstance(other, Sentence) \
               and self.pattern == other.pattern \
               and self.conditions == other.conditions \
               and self.result == other.result \
               and self.block == other.block

    def __str__(self):
        result_str = "\t" + str(self.pattern)
        if self.conditions:
            result_str += (", " + ",\t".join(list(map(str, self.conditions[::-1]))))
        if self.block:
            result_str += (", " + str(self.result))
        if self.block:
            result_str += (" :\n\t{\n\t\t" + ";\n\t\t".join(list(map(str, self.block))) + ";\n\t}")
        if self.result is not None and not self.block:
            result_str += (" = " + str(self.result))
        return result_str


class Condition(object):

    def __init__(self, result, pattern):
        self.result = result
        self.pattern = pattern

    def __str__(self):
        return str(self.result) + " : " + str(self.pattern)


class Expression(object):

    def __init__(self, terms):
        self.terms = terms

    def __eq__(self, other):
        return isinstance(other, Expression) and self.terms == other.terms

    def __hash__(self):
        return hash(self.__str__())

    def __str__(self):
        return " ".join(list(map(str, self.terms)))


class Term(ABC):

    def __init__(self, value):
        self.value = value

    @abstractmethod
    def __str__(self):
        return str(self.value)


class Char(Term):

    def __init__(self, value):
        super(Char, self).__init__(value)

    def __eq__(self, other):
        return isinstance(other, Char) and self.value == other.value

    def __str__(self):
        return super(Char, self).__str__()


class Macrodigit(Term):

    def __init__(self, value):
        super(Macrodigit, self).__init__(value)

    def __eq__(self, other):
        return isinstance(other, Macrodigit) and self.value == other.value

    def __str__(self):
        return super(Macrodigit, self).__str__()


class CompoundSymbol(Term):

    def __init__(self, value):
        super(CompoundSymbol, self).__init__(value)

    def __eq__(self, other):
        return isinstance(other, CompoundSymbol) and self.value == other.value

    def __str__(self):
        return super(CompoundSymbol, self).__str__()


class StructuralBrackets(Term):

    def __init__(self, value):
        super(StructuralBrackets, self).__init__(value)

    def __eq__(self, other):
        return isinstance(other, StructuralBrackets) and self.value == other.value

    def __str__(self):
        return "(" + " ".join(list(map(str, self.value))) + ")"


class CallBrackets(Term):

    def __init__(self, func_name, pos, content):
        super(CallBrackets, self).__init__(func_name)
        self.pos = pos
        self.content = content

    def __str__(self):
        return "<" + self.value + " " + " ".join(list(map(str, self.content))) + ">"


class Type(Enum):
    s = 1
    t = 2
    e = 3


class Variable(Term):

    def __init__(self, value, type_variable, pos, index=-1, sentence_index=-1):
        super(Variable, self).__init__(value)
        self.type_variable = type_variable
        self.index = index
        self.pos = pos
        self.sentence_index = sentence_index

    def __eq__(self, other):
        return isinstance(other, Variable) and self.type_variable == other.type_variable and self.index == other.index

    def __str__(self):
        if self.type_variable == Type.s:
            return "s." + str(self.index)
        elif self.type_variable == Type.t:
            return "t." + str(self.index)
        elif self.type_variable == Type.e:
            return "e." + str(self.index)



#! python -v
# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod
from enum import Enum


class DomainTag(Enum):
    Keyword = 1
    Ident = 2
    Number = 3
    Variable = 4
    Composite_symbol = 5
    Characters = 6
    Mark_sign = 7
    Left_bracket = 8
    Eop = 9
    Unknown = 10


class Token(ABC):

    def __init__(self, tag, coords, value):
        self.tag = tag
        self.coords = coords
        self.value = value

    def __eq__(self, other):
        return self.tag == other.tag and self.value == other.value

    @abstractmethod
    def __str__(self):
        return str(self.coords)


class KeywordToken(Token):

    def __init__(self, value, coords=None):
        super(KeywordToken, self).__init__(DomainTag.Keyword, coords, value)

    def __str__(self):
        return "Extern " + super(KeywordToken, self).__str__() + ": " + self.value


class IdentToken(Token):

    def __init__(self, value, coords=None):
        super(IdentToken, self).__init__(DomainTag.Ident, coords, value)

    def __str__(self):
        return "Ident " + super(IdentToken, self).__str__() + ": " + self.value


class VariableToken(Token):

    def __init__(self, value, coords=None):
        super(VariableToken, self).__init__(DomainTag.Variable, coords, value)

    def __str__(self):
        return "Variable " + super(VariableToken, self).__str__() + ": " + self.value


class CompositeSymbolToken(Token):

    def __init__(self, value, coords=None):
        super(CompositeSymbolToken, self).__init__(DomainTag.Composite_symbol, coords, value)

    def __str__(self):
        return "CompositeSymbol " + super(CompositeSymbolToken, self).__str__() + ": " + self.value


class CharacterToken(Token):

    def __init__(self, value, coords=None):
        super(CharacterToken, self).__init__(DomainTag.Characters, coords, value)

    def __str__(self):
        return "Characters " + super(CharacterToken, self).__str__() + ": " + self.value


class MarkSignToken(Token):

    def __init__(self, value, coords=None):
        super(MarkSignToken, self).__init__(DomainTag.Mark_sign, coords, value)

    def __str__(self):
        return "Sign " + super(MarkSignToken, self).__str__() + ": " + self.value


class LeftBracketToken(Token):
    def __init__(self, value, coords=None):
        super(LeftBracketToken, self).__init__(DomainTag.Left_bracket, coords, value)

    def __str__(self):
        return "Left Bracket: " + super(LeftBracketToken, self).__str__() + ": " + self.value


class NumberToken(Token):

    def __init__(self, value, coords=None):
        super(NumberToken, self).__init__(DomainTag.Number, coords, value)

    def __str__(self):
        return "Number " + super(NumberToken, self).__str__() + ": " + str(self.value)


class UnknownToken(Token):

    def __init__(self, value, coords=None):
        super(UnknownToken, self).__init__(DomainTag.Unknown, coords, value)

    def __str__(self):
        return super(UnknownToken, self).__str__() + "= " + self.value


class EopToken(Token):

    def __init__(self, coords=None):
        super(EopToken, self).__init__(DomainTag.Eop, coords, "")

    def __str__(self):
        return ""



#! python -v
# -*- coding: utf-8 -*-

from os.path import *

# Switch on/off debug information for program
DEBUG_MODE = False

# If DEBUG_MODE = True, then include directory, where contains refal program
TEST_DIRECTORY = join(dirname(dirname(__file__)), 'tests', 'files').replace("\\", "/")

# List of default Refal-functions
DEFAULT_FUNCTIONS = ["Mu", "Add", "Arg", "Br", "Card", "Chr", "Cp", "Dg", "Dgall", "Div", "Divmod",
                     "Explode", "First", "Get", "Implode", "Last", "Lenw", "Lower", "Mod", "Mul",
                     "Numb", "Open", "Ord", "Print", "Prout", "Put", "Putout", "Rp", "Step",
                     "Sub", "Symb", "Time", "Type", "Upper", "Freeze", "Dn",
                     "Up", "Ev-met", "Residue", "System", "Exit", "Close", "ExistFile", "GetCurrentDirectory",
                     "RemoveFile", "SizeOf", "GetPID", "GetPPID",
                     "Implode_Ext", "Explode_Ext", "TimeElapsed", "Compare", "Random", "RandomDigit", "Write",
                     "ListOfBuiltin", "Sysfun", "DeSysfun", "XMLParse", "GetEnv"]

# Delimiter when needs to divide information
LINE_DELIMITER = "=================="



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
import os
import sys
from shutil import rmtree
from refal import __version__

from setuptools import find_packages, setup, Command
from setuptools.command.test import test as command_test


NAME = 'refalchecker'
DESCRIPTION = 'Calculating/Verifying format of functions in Refal-5'
URL = 'https://github.com/runnerpeople/Refal5'
EMAIL = 'gosha8352@gmail.com'
AUTHOR = 'George Ivanov'
REQUIRES_PYTHON = '>=3.6.0'
VERSION = '.'.join(__version__.__version__)

REQUIRED = [
    'argparse'
]

EXTRAS = {}

here = os.path.abspath(os.path.dirname(__file__))

try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION


class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))

        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(VERSION))
        os.system('git push --tags')

        sys.exit()


class TestCommand(command_test):
    user_options = [("pytest-args=", "a", "Arguments to pass to pytest")]

    def initialize_options(self):
        command_test.initialize_options(self)
        self.pytest_args = ""

    def run_tests(self):
        import shlex

        # import here, cause outside the eggs aren't loaded
        import pytest

        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(),
    include_package_data=True,

    install_requires=REQUIRED,
    extras_require=EXTRAS,
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand,
        'test': TestCommand
    },
    entry_points={
        'console_scripts': ['refalcheck=refal.refalcheck:main'],
    },
    tests_require=["pytest"],
    test_suite='tests'
)



#! python -v
# -*- coding: utf-8 -*-


class Fragment(object):

    def __init__(self, starting, following):
        self.starting = starting
        self.following = following

    def __str__(self):
        return str(self.starting) + "-" + str(self.following)


class Position(object):

    def __init__(self, text, line=1, pos=1, index=0):
        self.text = text
        self.line = line
        self.pos = pos
        self.index = index

    def __eq__(self, other):
        return self.index == other.index

    def __lt__(self, other):
        return self.index < other.index

    def __str__(self):
        return "(" + str(self.line) + "," + str(self.pos) + ")"

    def is_eof(self):
        return self.index == len(self.text)

    def cp(self):
        return -1 if self.index == len(self.text) else ord(self.text[self.index])

    def letter(self):
        return "" if self.index >= len(self.text) else self.text[self.index]

    def read(self, n):
        cur = Position(self.text, self.line, self.pos, self.index)
        result_str = ""
        for i in range(0, n):
            if cur is not None:
                result_str += cur.letter()
                cur = next(cur, None)
        return result_str

    def __iter__(self):
        return self

    def is_white_space(self):
        return self.index != len(self.text) and self.letter().isspace()

    def is_letter(self):
        return self.index != len(self.text) and self.letter().isalpha()

    def is_digit(self):
        return self.index != len(self.text) and self.letter().isdigit()

    def is_letter_or_digit(self):
        return self.index != len(self.text) and self.letter().isalnum()

    def is_decimal_digit(self):
        return self.index != len(self.text) and "0" <= self.text[self.index] <= "9"

    def is_latin_letter(self):
        return self.index != len(self.text) and ("a" <= self.text[self.index] <= "z" or
                                                 "A" <= self.text[self.index] <= "Z")

    def is_new_line(self):
        if self.index == len(self.text):
            return True
        if self.text[self.index] == "\r" and self.index + 1 < len(self.text):
            return self.text[self.index+1] == "\n"
        return self.text[self.index] == "\n"

    def __next__(self):
        if not self.has_next():
            raise StopIteration
        elif self.is_new_line():
            if self.text[self.index] == "\r":
                self.index += 1
            self.line += 1
            self.pos = 1
        else:
            self.pos += 1
        self.index += 1
        return self

    def has_next(self):
        return self.index < len(self.text)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import re

import pytest

from refal.refalcheck import *


@pytest.fixture
def built_in_parser():
    file_type_path = join(dirname(dirname(__file__)), 'refal', 'built-in.type').replace("\\", "/")
    parser_refal_type = calculate_ast(file_type_path, True)

    return parser_refal_type


def file_parser(name_file, is_file_type=False):
    file_refal_path = join(dirname(__file__), 'files', name_file).replace("\\", "/")
    parser_refal = calculate_ast(file_refal_path, is_file_type)

    return parser_refal


def run_easy_test(name_file, built_in_parser, contain_file_type=False):
    parser_refal = file_parser(name_file)
    parser_refal_type = built_in_parser

    if contain_file_type:
        append_ast(parser_refal_type, calculate_ast(join(dirname(__file__), 'files', name_file).replace("\\", "/").
                                                    replace(".ref", ".type"), True))

    parser_refal.semantics_call(parser_refal_type.ast)

    assert not parser_refal.isError
    assert not parser_refal_type.isError

    if not parser_refal.isError and not parser_refal_type.isError:
        calculation = Calculation(parser_refal.ast, parser_refal_type.ast)
        calculation.calculate_system()
        output = calculation.print_format_function()

        output_format = [line for line in output.split("\n") if line and line != '/* result program refalcheck */']

        return output_format


@pytest.mark.timeout(1)
def test_1(built_in_parser):
    output_format = run_easy_test("test1.ref", built_in_parser)

    assert output_format[0] == "*Trans-line e = e"
    assert output_format[1] == "*Table  = (('cane') 'dog') (('gatto') 'cat') (('cavallo') 'horse') (('rana') " \
                               "'frog') (('porco') 'pig')"
    assert output_format[2] == "*Trans (e) e = e"


@pytest.mark.timeout(1)
def test_3(built_in_parser):
    output_format = run_easy_test("test3.ref", built_in_parser)

    assert output_format[0] == "*F s e = s"


@pytest.mark.timeout(10)
def test_4(built_in_parser):
    output_format = run_easy_test("test4.ref", built_in_parser)

    assert output_format[0] == "Go  = "
    assert output_format[1] == "*CheckArg e = "
    assert output_format[2] == "*ReadLines e = e"
    assert output_format[3] == "*Format e = '<html><body>\\n<h1>Оглавление</h1>\\n' e '</body></html>'"
    assert output_format[4] == "*Trim-Left e = e"
    assert output_format[5] == "*FormatRec s e = e"
    assert output_format[6] == "*AddTableContent s e = e"


@pytest.mark.timeout(1)
def test_5(built_in_parser):
    output_format = run_easy_test("test5.ref", built_in_parser)

    assert output_format[0] == "Go  = "
    assert output_format[1] == "*CheckArg e = "
    assert output_format[2] == "*ReadLines e = e"
    assert output_format[3] == "*Format e = '<html><body>\\n' e '</body></html>'"
    assert output_format[4] == "*Trim-Left e = e"
    assert output_format[5] == "*FormatRec e = e"
    assert output_format[6] == "*CodeBlock e = e"
    assert output_format[7] == "*DeleteCodeBlock e = e"
    assert output_format[8] == "*Fix e = e"


@pytest.mark.timeout(60)
def ttest_6(built_in_parser):
    output_format = run_easy_test("test6.ref", built_in_parser)

    assert output_format[0] == "Go  = "
    assert output_format[1] == "*CheckArg e = "
    assert output_format[2] == "*CountWords s e = "
    assert output_format[3] == "*CountWordsRec s e = "
    assert output_format[4] == "*Trim-Left e = e"
    assert output_format[5] == "*DoCountWordsFile (s) e = e"
    assert output_format[6] == "*DoCountWordsInput (s) e = e"


@pytest.mark.timeout(1)
def test_7(built_in_parser):
    output_format = run_easy_test("test7.ref", built_in_parser)

    assert output_format[0] == "Go  = "
    assert output_format[1] == "*ReadInput e = e"
    assert output_format[2] == "*CheckNumber s (e) = s s s s e '\\n'"


@pytest.mark.timeout(1)
def test_8(built_in_parser):
    output_format = run_easy_test("test8.ref", built_in_parser)

    assert output_format[0] == "*DoHexDigit s e s = e"


@pytest.mark.timeout(1)
def test_10(built_in_parser):
    output_format = run_easy_test("test10.ref", built_in_parser)

    assert output_format[0] == "*F s e (e) = s e"
    assert output_format[1] == "*G s e = s"
    assert output_format[2] == "*H e (e) = e"


@pytest.mark.timeout(1)
def test_11(built_in_parser):
    output_format = run_easy_test("test11.ref", built_in_parser)

    assert output_format[0] == "*Job e = e"
    assert output_format[1] == "*Trans (e) e = e"
    assert output_format[2] == "*Fact s = s e"
    assert output_format[3] == "*Table  = (('cane') 'dog') (('gatto') 'cat') (('cavallo') 'horse') (('rana') 'frog') " \
                               "(('porco') 'pig')"
    assert output_format[4] == "*Trans-line e = e"


@pytest.mark.timeout(1)
def test_12(built_in_parser):
    output_format = run_easy_test("test12.ref", built_in_parser)

    assert output_format[0] == "*F @ = @"


@pytest.mark.timeout(1)
def test_13(built_in_parser):
    output_format = run_easy_test("test13.ref", built_in_parser)

    assert output_format[0] == "*Fab (e) e = e"


@pytest.mark.timeout(1)
def test_14(built_in_parser):
    output_format = run_easy_test("test14.ref", built_in_parser, True)

    assert output_format[0] == "*F A e = A"


@pytest.mark.timeout(1)
def test_15(built_in_parser, capsys):
    _ = run_easy_test("test15.ref", built_in_parser)

    captured = capsys.readouterr()
    output = re.sub(r'\.[\d]+', '', captured.err)
    assert output == "Function F, sentence 0, there isn't solution for equation: A A e : in[F] => A\n"


def ttest_17(built_in_parser):
    output_format = run_easy_test("test17.ref", built_in_parser)

    assert output_format[0] == "Apply t e = e"
    assert output_format[1] == "Map t e = e"
    assert output_format[2] == "Reduce t t e = t"
    assert output_format[3] == "MapAccum t t e = t e"
    assert output_format[4] == "*DoMapAccum t t (e) e = t e"
    assert output_format[5] == "*MapAccum-AddScanned t e (e) = t (e)"
    assert output_format[6] == "DelAccumulator t e = e"
    assert output_format[7] == "*LOAD-SAVE-HANDLE  = 39"
    assert output_format[8] == "LoadFile e = e"
    assert output_format[9] == "*DoLoadFile e = e"
    assert output_format[10] == "SaveFile (e) e = e"
    assert output_format[11] == "*SaveFile-WriteBracketLine (e) = "
    assert output_format[12] == "Inc s = s e"
    assert output_format[13] == "Dec s = s e"
    assert output_format[14] == "ArgList  = e"
    assert output_format[15] == "*DoArgList s = e"
    assert output_format[16] == "*SwDoArgList s e = e"
    assert output_format[17] == "Trim e = e"
    assert output_format[18] == "*Trim-R e = e"


@pytest.mark.timeout(1)
def test_18(built_in_parser):
    output_format = run_easy_test("test18.ref", built_in_parser, True)

    assert output_format[0] == "generator_GenSentence t = () ('  ' s s s s e) e"
    assert output_format[1] == "*SkipIndentAccum ('    ') e = e"
    assert output_format[2] == "*CompileSentence (e) (e) = e"


@pytest.mark.timeout(1)
def test_19(built_in_parser):
    output_format = run_easy_test("test19.ref", built_in_parser)

    assert output_format[0] == "refal05c_WriteError e ((s s) e) = "
    assert output_format[1] == "*StrFromSrcPos (s s) = e"


@pytest.mark.timeout(1)
def test_20(built_in_parser):
    output_format = run_easy_test("test20.ref", built_in_parser, True)

    assert output_format[0] == "refal05c_ProcessEachSource (s e) = e t"
    assert output_format[1] == "*CompileSource-SwSuccessedParse (e) (e) s e = e t"


def te_21(built_in_parser):
    output_format = run_easy_test("test21.ref", built_in_parser, True)

    assert output_format[0] == "R05-Generate-ToFile (e) e = "
    assert output_format[1] == "R05-Generate-ToLines e = e"
    assert output_format[2] == "Generate (e) e = e"
    assert output_format[3] == "R05-Generate-Aux (e) (e) e = e"
    assert output_format[4] == "generator_GenTreeItem (e) (s e) = (e) e"
    assert output_format[5] == "TextFromScopeClass s = e"
    assert output_format[6] == "GenDeclaration s e = (s s s s s s ' struct r05_function r05f_' e ';')"
    assert output_format[7] == "GenFunction s (e) s e = (e s s s s) (e) e"
    assert output_format[8] == "GenFunctionBody s e = (s e s) e"
    assert output_format[9] == "AddFailCommand e = e t"
    assert output_format[10] == "generator_GenSentence t = () ('  ' s s s s e) e"
    assert output_format[11] == "SkipIndentAccum ('    ') e = e"
    assert output_format[12] == "RangeVar-B s = 'bb[' e ']'"
    assert output_format[13] == "RangeVar-E s = 'be[' e ']'"
    assert output_format[14] == "RangeVarsPtr s = 'bb+' e"
    assert output_format[15] == "SafeComment e = e"
    assert output_format[16] == "generator_GenCommand (e) (s e) = (e) e"
    assert output_format[17] == "CmdComment (e) e = (e ' */')"
    assert output_format[18] == "CmdDeclareVar (e) s s e = (e ';') e"
    assert output_format[19] == "CmdRangeArray (e) s = (e '] = { 0 };') (e '] = { 0 };')"
    assert output_format[20] == "CmdInitB0 (e) = (e ', arg_begin, arg_end);')"
    assert output_format[21] == "CmdMatch (e) s s s e = (e '))') (e '  continue;')"
    assert output_format[22] == "MatchFunc s s e = s s s s e"
    assert output_format[23] == "SymbolFunc s = s s s s e"
    assert output_format[24] == "StrFromDirection s = s s s e 't'"
    assert output_format[25] == "MatchArgs s s e = e"
    assert output_format[26] == "SymbolTextRep s t e = e"
    assert output_format[27] == "CmdEmpty (e) s = (e ']))') (e '  continue;')"
    assert output_format[28] == "CmdClosedE (e) s 'e' e = (e '];') (e '];')"
    assert output_format[29] == "CmdOpenedE-Start (e) s 'e' e = (e ' = 0;') (e ' = 0;') (e 'r05_start_e_loop();') (e " \
                                "'do {')"
    assert output_format[30] == "CmdOpenedE-End (e) s 'e' e = (e '));') (e 'r05_stop_e_loop();')"
    assert output_format[31] == "CmdSave (e) s s = (e '];') (e '];')"
    assert output_format[32] == "CmdEmptyResult (e) = () (e 'r05_reset_allocator();')"
    assert output_format[33] == "CmdResultArray (e) s = (e '] = { 0 };')"
    assert output_format[34] == "CmdAllocateElem (e) s e = (e)"
    assert output_format[35] == "ElSymbol s t e = s s s s e ');'"
    assert output_format[36] == "ElString s e = 'chars(\"' e ');'"
    assert output_format[37] == "ElOpenBracket s = 'open_bracket(n+' e ');'"
    assert output_format[38] == "ElCloseBracket s = 'close_bracket(n+' e ');'"
    assert output_format[39] == "ElOpenCall s = 'open_call(n+' e ');'"
    assert output_format[40] == "ElCloseCall s = 'close_call(n+' e ');'"
    assert output_format[41] == "ElSavePos s = 'insert_pos(n+' e ');'"
    assert output_format[42] == "ElVariable s e = s 'var(' s e ');'"
    assert output_format[43] == "CmdLinkBrackets (e) s s = (e ']);')"
    assert output_format[44] == "CmdPushStack (e) s = (e ']);')"
    assert output_format[45] == "CmdInsertVar (e) s s s e = (e ');')"
    assert output_format[46] == "CmdReturnResult (e) = (e 'r05_splice_from_freelist(arg_begin);') (e " \
                                "'r05_splice_to_freelist(arg_begin, arg_end);') (e 'return;')"
    assert output_format[47] == "EscapeString e = e"
    assert output_format[48] == "EscapeChar s = s e"
    assert output_format[49] == "EscapeChar-Aux s s = s e"
    assert output_format[50] == "EscapeChar-SwCompare s s s s = s e"
    assert output_format[51] == "Var s s e = s e"
    assert output_format[52] == "VarPtr s s e = '&' s e"
    assert output_format[53] == "EVar-B s 'e' e = 'e' e"
    assert output_format[54] == "EVar-E s 'e' e = 'e' e"
    assert output_format[55] == "Elem s = 'n[' e ']'"
    assert output_format[56] == "ElemPtr s = 'n+' e"
    assert output_format[57] == "GenNative (e s) e = (BeginNative e s) e (EndNative)"
    assert output_format[58] == "GenPostprocess (e) e = e"
    assert output_format[59] == "generator_EnumerateLines (e) s (e) = s (e)"
    assert output_format[60] == "LineDirective s e = ('#line ' e '\"')"
    assert output_format[61] == "CompileSentence (e) (e) = e"
    assert output_format[62] == "CompileSentence-Aux s (e) (e) e = e"
    assert output_format[63] == "GenPattern e = s (e) (e)"
    assert output_format[64] == "DoGenPattern s e (e) (e) = s (e) (e)"
    assert output_format[65] == "SaveRanges s (e) (e) (e) e = s e (e) (e)"
    assert output_format[66] == "GenResult (e) e = s (e) (e)"
    assert output_format[67] == "DecUsings s s e = e"
    assert output_format[68] == "ComposeSentenceCommands s (e) (e) s (e) (e) = e"
    assert output_format[69] == "FilterCommonVarsAndPatternCommands (e) (e) e = (e) e"
    assert output_format[70] == "ComposeSentenceCommands-Aux s s (e) (e) e = e"
    assert output_format[71] == "generator_MakeDeclaration (s s e) = (CmdDeclareVar s s e) e"
    assert output_format[72] == "MakeDeclaration-Aux s s e = e"
    assert output_format[73] == "MakeCmdResultCommand s = e"
    assert output_format[74] == "GenerateResult-OpenELoops e = e"


def ttest_22(built_in_parser):
    output_format = run_easy_test("test22.ref", built_in_parser, True)

    assert output_format[0] == "R05-Parse-File e = s e"
    assert output_format[1] == "R05-Parse-String e = s e"
    assert output_format[2] == "R05-Parse-Tokens e = s e"
    assert output_format[3] == "R05-Parse-Aux t e = s e"
    assert output_format[4] == "R05-Parse-SwErrors (e) e = s e"
    assert output_format[5] == "SortErrors e = e"
    assert output_format[6] == "DoSortErrors (e) e = e"
    assert output_format[7] == "SortErrors-Insert e t = e t"
    assert output_format[8] == "SortErrors-Insert-Compare e t t s = e t t"
    assert output_format[9] == "CompareError ((s s) e) ((s s) e) = s"
    assert output_format[10] == "ParseProgram e t = t (e) e"
    assert output_format[11] == "ParseElement (s t e) e t = e t"
    assert output_format[12] == "parser_MakeListFunction s (t e) = (Function t s (e) Sentences)"
    assert output_format[13] == "ParseList (s t e) e t = (s t e) e (ErrorList e)"
    assert output_format[14] == "ParseListKeyWord (s t e) e t = (ListKeyWord s) e t"
    assert output_format[15] == "ParseFunction e t = t e t"
    assert output_format[16] == "ParseScope e t = (Scope s) e t"
    assert output_format[17] == "ParseBody e t = (Body (e) s e) e t"
    assert output_format[18] == "ParseSentences e t = (Sentences (e) e) (TkCloseBlock t) e t"
    assert output_format[19] == "ParseSentence e t = (Sentence (e) (e) (e)) e t"
    assert output_format[20] == "ParsePattern s e t = (Pattern (e) (e) e) e t"
    assert output_format[21] == "CleanupDoubleVars e (e) = e"
    assert output_format[22] == "ParsePatternTerm e t = (s e) e t"
    assert output_format[23] == "ParseResult s e (e) t = (Result (e) e) e t"
    assert output_format[24] == "ParseResultTerm e (e) t = (s e) e (e) t"
    assert output_format[25] == "BracketPairName s = s"
    assert output_format[26] == "SemanticCheck t (e) e = t e"
    assert output_format[27] == "FindOneEntry t (e) e = t e"
    assert output_format[28] == "LoadBuiltins  = e"
    assert output_format[29] == "parser_BuiltinDeclaration (s s s) = e"
    assert output_format[30] == "Normalize e = e"
    assert output_format[31] == "CheckRepeatedDefinitions t e (e) = t e"
    assert output_format[32] == "RemoveReference (e) e = e"
    assert output_format[33] == "parser_AddUnresolved (ErrorList e) (t e) = (ErrorList e (t e))"
    assert output_format[34] == "EL-Create  = (ErrorList)"
    assert output_format[35] == "EL-AddErrorAt (ErrorList e) t e = (ErrorList e (t e))"
    assert output_format[36] == "EL-AddUnexpected t (s t e) e = (ErrorList e (t e))"
    assert output_format[37] == "EL-Destroy (ErrorList e) = e"


def ttest_23(built_in_parser):
    output_format = run_easy_test("test23.ref", built_in_parser)

    assert output_format[0] == "R05-TextFromTree e = e"
    assert output_format[1] == "Extern e = '$EXTERN ' e ';\\n'"
    assert output_format[2] == "Function s (e) t e = e s '\\n'"
    assert output_format[3] == "Entry  = '$ENTRY '"
    assert output_format[4] == "Local  = "
    assert output_format[5] == "Sentences e = e"
    assert output_format[6] == "*TextFromSentence ((e) (e)) = '  ' e ';\\n'"
    assert output_format[7] == "Native (e s) e = '* file: \"' e '%%\\n'"
    assert output_format[8] == "*FlatLines e = e"
    assert output_format[9] == "Symbol s e = e"
    assert output_format[10] == "Char s = '\\' s e '\\'"
    assert output_format[11] == "Number s = e"
    assert output_format[12] == "Name t e = e"
    assert output_format[13] == "Variable s e = s '.' e"
    assert output_format[14] == "Brackets e = '(' e ')'"
    assert output_format[15] == "CallBrackets e = '<' e '>'"
    assert output_format[16] == "TextFromExpr e = e"
    assert output_format[17] == "*TextFromExpr-Char e = s e"
    assert output_format[18] == "*TextFromTerm (s e) = e"
    assert output_format[19] == "EscapeChar s = s e"
    assert output_format[20] == "*EscapeChar-Aux s s = s e"
    assert output_format[21] == "*EscapeChar-SwCompare s s s s = s e"
    assert output_format[22] == "*CharFromHex s = s"


@pytest.mark.timeout(1)
def test_24(built_in_parser):
    output_format = run_easy_test("test24.ref", built_in_parser)

    assert output_format[0] == "*F s s s s = s"
    assert output_format[1] == "*G s s s s = s"
    assert output_format[2] == "*H s s e = s"


@pytest.mark.timeout(1)
def test_25(built_in_parser, capsys):
    _ = run_easy_test("test25.ref", built_in_parser)

    captured = capsys.readouterr()
    output = re.sub(r'\.[\d]+', '', captured.err)
    assert output == "Function F, sentence 0, there isn't solution for equation: 'a' () 'z' 'z' 'z' 'z' : in[F] => s s () e\n"


@pytest.mark.timeout(1)
def test_26(built_in_parser):
    output_format = run_easy_test("test26.ref", built_in_parser)

    assert output_format[0] == "*EL-Create  = (ErrorList)"
    assert output_format[1] == "*EL-AddErrorAt (ErrorList e) t e = (ErrorList e (t e))"
    assert output_format[2] == "*EL-AddUnexpected (ErrorList e) (s t e) e = (ErrorList e (t e))"
    assert output_format[3] == "*EL-Destroy (ErrorList e) = e"
    assert output_format[4] == "StrFromToken s e = s e s"


@pytest.mark.timeout(1)
def test_27(built_in_parser):
    output_format = run_easy_test("test27.ref", built_in_parser)

    assert output_format[0] == "*F  = 1 2 3"
    assert output_format[1] == "*G 1 2 3 4 5 = "
    assert output_format[2] == "H 4 5 = "


@pytest.mark.timeout(1)
def test_28(built_in_parser):
    output_format = run_easy_test("test28.ref", built_in_parser)

    assert output_format[0] == "*F e s = "
    assert output_format[1] == "*G (e) e = "
    assert output_format[2] == "*H (e) e s = "


@pytest.mark.timeout(1)
def test_29(built_in_parser):
    output_format = run_easy_test("test29.ref", built_in_parser)

    assert output_format[0] == "*F t = "
    assert output_format[1] == "*F2 t = "
    assert output_format[2] == "*F3 t = "


@pytest.mark.timeout(1)
def test_30(built_in_parser):
    output_format = run_easy_test("test30.ref", built_in_parser)

    assert output_format[0] == "*G (e) e = e"
    assert output_format[1] == "*F s e = s"


@pytest.mark.timeout(1)
def test_31(built_in_parser):
    output_format = run_easy_test("test31.ref", built_in_parser)

    assert output_format[0] == "*G (e) e = e"
    assert output_format[1] == "*F s e = s"


@pytest.mark.timeout(1)
def test_32(built_in_parser):
    output_format = run_easy_test("test32.ref", built_in_parser)

    assert output_format[0] == "*F t e = t"
    assert output_format[1] == "*G e t = t"
    assert output_format[2] == "*H e t = e"


@pytest.mark.timeout(1)
def test_33(built_in_parser):
    output_format = run_easy_test("test33.ref", built_in_parser)

    assert output_format[0] == "*F (((t))) = "
    assert output_format[1] == "*F2 ((t)) = "
    assert output_format[2] == "*F3 (((t))) = "


@pytest.mark.timeout(1)
def test_34(built_in_parser):
    output_format = run_easy_test("test34.ref", built_in_parser)

    assert output_format[0] == "*F I I I e = "
    assert output_format[1] == "*F2 I I e = "
    assert output_format[2] == "*F3 I I I e = "


@pytest.mark.timeout(1)
def test_35(built_in_parser):
    output_format = run_easy_test("test35.ref", built_in_parser)

    assert output_format[0] == "*Derivative t = t"
    assert output_format[1] == "*Opt t = t"
    assert output_format[2] == "*Opt-Rules t s t = t"
    assert output_format[3] == "DerivOpt t = t"


